# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from sklearn.linear_model import (
    LogisticRegression,
    RidgeClassifier,
    RidgeClassifierCV,
)
from sklearn.svm import LinearSVC
from ..common._apply_operation import (
    apply_cast,
    apply_add,
    apply_sigmoid,
    apply_softmax,
    apply_normalizer,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import BooleanTensorType, DoubleTensorType, guess_proto_type
from ..common.utils_classifier import get_label_classes, _finalize_converter_classes
from ..proto import onnx_proto


def apply_logistic(scope, input_name, output_name, container, proto_dtype):
    sig_name = scope.get_unique_variable_name(input_name + "sig")
    apply_sigmoid(scope, input_name, sig_name, container)
    apply_normalizer(
        scope,
        sig_name,
        output_name,
        container,
        norm="L1",
        use_float=proto_dtype == onnx_proto.TensorProto.FLOAT,
    )


def convert_sklearn_linear_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    coefficients = op.coef_.flatten().astype(float).tolist()
    classes = get_label_classes(scope, op)
    number_of_classes = len(classes)
    use_linear_op = container.is_allowed({"LinearClassifier"})

    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]

    if isinstance(op.intercept_, (float, np.float32)) and op.intercept_ == 0:
        # fit_intercept = False
        intercepts = [0.0] * number_of_classes if number_of_classes != 2 else [0.0]
    else:
        intercepts = op.intercept_.tolist()

    if number_of_classes == 2:
        coefficients = list(map(lambda x: -1 * x, coefficients)) + coefficients
        intercepts = list(map(lambda x: -1 * x, intercepts)) + intercepts

    multi_class = 0
    if hasattr(op, "multi_class"):
        if op.multi_class == "ovr":
            multi_class = 1
        else:
            multi_class = 2

    classifier_type = "LinearClassifier"
    classifier_attrs = {"name": scope.get_unique_operator_name(classifier_type)}

    classifier_attrs["coefficients"] = coefficients
    classifier_attrs["intercepts"] = intercepts
    classifier_attrs["multi_class"] = 1 if multi_class == 2 else 0
    if use_raw_scores or isinstance(
        op, (LinearSVC, RidgeClassifier, RidgeClassifierCV)
    ):
        classifier_attrs["post_transform"] = "NONE"
    elif isinstance(op, LogisticRegression):
        ovr = op.multi_class in ["ovr", "warn"] or (
            op.multi_class == "auto"
            and (op.classes_.size <= 2 or op.solver == "liblinear")
        )
        classifier_attrs["post_transform"] = "LOGISTIC" if ovr else "SOFTMAX"
    else:
        classifier_attrs["post_transform"] = (
            "LOGISTIC" if multi_class > 2 else "SOFTMAX"
        )

    if all(isinstance(i, str) for i in classes):
        class_labels = [str(i) for i in classes]
        classifier_attrs["classlabels_strings"] = class_labels
    elif all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
        class_labels = [int(i) for i in classes]
        classifier_attrs["classlabels_ints"] = class_labels
    else:
        raise RuntimeError("Label vector must be a string or a integer " "tensor.")

    if not use_linear_op or type(operator.inputs[0].type) in (DoubleTensorType,):
        # Double -> double parameters not supported in ONNX LinearClassifier
        proto_dtype = guess_proto_type(operator.inputs[0].type)
        coef = scope.get_unique_variable_name("coef")
        model_coef = np.array(classifier_attrs["coefficients"], dtype=np.float64)
        model_coef = model_coef.reshape((number_of_classes, -1)).T
        container.add_initializer(
            coef, proto_dtype, model_coef.shape, model_coef.ravel().tolist()
        )
        intercept = scope.get_unique_variable_name("intercept")
        model_intercept = np.array(classifier_attrs["intercepts"], dtype=np.float64)
        model_intercept = model_intercept.reshape((number_of_classes, -1)).T
        container.add_initializer(
            intercept,
            proto_dtype,
            model_intercept.shape,
            model_intercept.ravel().tolist(),
        )
        multiplied = scope.get_unique_variable_name("multiplied")
        container.add_node(
            "MatMul",
            [operator.inputs[0].full_name, coef],
            multiplied,
            name=scope.get_unique_operator_name("MatMul"),
        )

        if use_raw_scores:
            raw_score_name = operator.outputs[1].full_name
        else:
            raw_score_name = scope.get_unique_variable_name("raw_scores")
        apply_add(scope, [multiplied, intercept], raw_score_name, container)

        argmax_output_name = scope.get_unique_variable_name("label")
        container.add_node(
            "ArgMax",
            raw_score_name,
            argmax_output_name,
            name=scope.get_unique_operator_name("ArgMax"),
            axis=1,
        )
        _finalize_converter_classes(
            scope,
            argmax_output_name,
            operator.outputs[0].full_name,
            container,
            np.array(class_labels),
            onnx_proto.TensorProto.DOUBLE,
        )

        if use_raw_scores:
            return

        if classifier_attrs["post_transform"] == "LOGISTIC":
            apply_logistic(
                scope,
                raw_score_name,
                operator.outputs[1].full_name,
                container,
                proto_dtype=onnx_proto.TensorProto.DOUBLE,
            )
            return
        elif classifier_attrs["post_transform"] == "SOFTMAX":
            apply_softmax(
                scope, raw_score_name, operator.outputs[1].full_name, container
            )
            return

        raise NotImplementedError(
            "post_transform '{}' is not supported with double.".format(
                classifier_attrs["post_transform"]
            )
        )

    label_name = operator.outputs[0].full_name
    input_name = operator.inputs[0].full_name
    if isinstance(operator.inputs[0].type, BooleanTensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope,
            input_name,
            cast_input_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        input_name = cast_input_name

    if use_raw_scores:
        container.add_node(
            classifier_type,
            input_name,
            [label_name, operator.outputs[1].full_name],
            op_domain="ai.onnx.ml",
            **classifier_attrs
        )
    elif (
        isinstance(op, (LinearSVC, RidgeClassifier, RidgeClassifierCV))
        and op.classes_.shape[0] <= 2
    ):
        raw_scores_tensor_name = scope.get_unique_variable_name("raw_scores_tensor")
        positive_class_index_name = scope.get_unique_variable_name(
            "positive_class_index"
        )

        container.add_initializer(
            positive_class_index_name, onnx_proto.TensorProto.INT64, [], [1]
        )

        if (
            hasattr(op, "_label_binarizer")
            and op._label_binarizer.y_type_ == "multilabel-indicator"
        ):
            y_pred_name = scope.get_unique_variable_name("y_pred")
            binarised_label_name = scope.get_unique_variable_name("binarised_label")

            container.add_node(
                classifier_type,
                input_name,
                [y_pred_name, raw_scores_tensor_name],
                op_domain="ai.onnx.ml",
                **classifier_attrs
            )
            container.add_node(
                "Binarizer",
                raw_scores_tensor_name,
                binarised_label_name,
                op_domain="ai.onnx.ml",
            )
            apply_cast(
                scope,
                binarised_label_name,
                label_name,
                container,
                to=onnx_proto.TensorProto.INT64,
            )
        else:
            container.add_node(
                classifier_type,
                input_name,
                [label_name, raw_scores_tensor_name],
                op_domain="ai.onnx.ml",
                **classifier_attrs
            )
        container.add_node(
            "ArrayFeatureExtractor",
            [raw_scores_tensor_name, positive_class_index_name],
            operator.outputs[1].full_name,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        )
    else:
        # Make sure the probability sum is 1 over all classes
        if multi_class > 0 and not isinstance(
            op, (LinearSVC, RidgeClassifier, RidgeClassifierCV)
        ):
            probability_tensor_name = scope.get_unique_variable_name(
                "probability_tensor"
            )
            container.add_node(
                classifier_type,
                input_name,
                [label_name, probability_tensor_name],
                op_domain="ai.onnx.ml",
                **classifier_attrs
            )
            use_float = type(operator.inputs[0].type) not in (DoubleTensorType,)
            apply_normalizer(
                scope,
                probability_tensor_name,
                operator.outputs[1].full_name,
                container,
                norm="L1",
                use_float=use_float,
            )
        elif (
            hasattr(op, "_label_binarizer")
            and op._label_binarizer.y_type_ == "multilabel-indicator"
        ):
            y_pred_name = scope.get_unique_variable_name("y_pred")
            binarised_label_name = scope.get_unique_variable_name("binarised_label")

            container.add_node(
                classifier_type,
                input_name,
                [y_pred_name, operator.outputs[1].full_name],
                op_domain="ai.onnx.ml",
                **classifier_attrs
            )
            container.add_node(
                "Binarizer",
                operator.outputs[1].full_name,
                binarised_label_name,
                op_domain="ai.onnx.ml",
            )
            apply_cast(
                scope,
                binarised_label_name,
                label_name,
                container,
                to=onnx_proto.TensorProto.INT64,
            )
        else:
            container.add_node(
                classifier_type,
                input_name,
                [label_name, operator.outputs[1].full_name],
                op_domain="ai.onnx.ml",
                **classifier_attrs
            )


register_converter(
    "SklearnLinearClassifier",
    convert_sklearn_linear_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)
register_converter(
    "SklearnLinearSVC",
    convert_sklearn_linear_classifier,
    options={
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from sklearn.svm import SVC, NuSVC
from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType
from ..common.data_types import StringTensorType
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_svm_output_shapes(operator):
    """
    For SVM classifiers, allowed input/output patterns are
        1. [N, C] ---> [N], A sequence of map
    Note that the second case is not allowed as long as ZipMap only
    produces dictionary.

    For SVM regressors, allowed input/output patterns are
        1. [N, C] ---> [N]

    For both of SVC and SVR, the inputs should numerical tensor(s).
    For SVC with batch size 1, the first output is the label and the
    second output is a map used to store all class probabilities (For
    a key-value pair, the value is assigned to the class specified by
    the key). If batch size is larger than 1, we need to use a sequence
    of maps to denote class probabilities. Regarding SVR, we just
    produce a scalar for each example. If there are N examples, the
    output shape would be [N, 1].
    """
    op = operator.raw_operator

    N = operator.inputs[0].get_first_dimension()
    if operator.type in ["SklearnOneClassSVM"]:
        operator.outputs[0].type = Int64TensorType([N, 1])
        operator.outputs[1].type.shape = [N, 1]
    elif operator.type in ["SklearnSVC"] or isinstance(op, (SVC, NuSVC)):
        number_of_classes = len(op.classes_)
        check_input_and_output_numbers(
            operator, input_count_range=[1, None], output_count_range=[1, 2]
        )

        if all(isinstance(i, str) for i in op.classes_):
            operator.outputs[0].type = StringTensorType([N])
            operator.outputs[1].type.shape = [N, number_of_classes]
        elif all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in op.classes_):
            operator.outputs[0].type = Int64TensorType([N])
            operator.outputs[1].type.shape = [N, number_of_classes]
        else:
            raise RuntimeError(
                "Class labels should be either all strings or "
                "all integers. C++ backends do not support "
                "mixed types."
            )

    elif operator.type in ["SklearnSVR"]:
        check_input_and_output_numbers(
            operator, input_count_range=[1, None], output_count_range=1
        )

        operator.outputs[0].type.shape = [N, 1]
    else:
        raise RuntimeError(
            "New kind of SVM, no shape calculator exist for '{}'.".format(operator.type)
        )


register_shape_calculator("SklearnOneClassSVM", calculate_sklearn_svm_output_shapes)
register_shape_calculator("SklearnSVC", calculate_sklearn_svm_output_shapes)
register_shape_calculator("SklearnSVR", calculate_sklearn_svm_output_shapes)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common.data_types import (
    Int64TensorType,
    StringTensorType,
    Int32TensorType,
    FloatTensorType,
    DoubleTensorType,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_sklearn_one_hot_encoder(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts *OneHotEncoder* into ONNX.
    It supports multiple inputs of types
    string or int64.
    """
    ohe_op = operator.raw_operator

    if len(operator.inputs) > 1:
        all_shapes = [inp.type.shape[1] for inp in operator.inputs]
        if any(map(lambda x: not isinstance(x, int) or x < 1, all_shapes)):
            raise RuntimeError(
                "Shapes must be known when OneHotEncoder is converted. "
                "There are {} inputs with the following number of columns "
                "{}.".format(len(operator.inputs), all_shapes)
            )
        total = sum(all_shapes)
        if total != len(ohe_op.categories_):
            raise RuntimeError(
                "Mismatch between the number of sets of categories {} and "
                "the total number of inputs columns {}.".format(
                    len(ohe_op.categories_), total
                )
            )

        enum_cats = []
        index_inputs = 0

        for index, cats in enumerate(ohe_op.categories_):
            while sum(all_shapes[: index_inputs + 1]) <= index:
                index_inputs += 1
            index_in_input = index - sum(all_shapes[:index_inputs])

            inp = operator.inputs[index_inputs]
            if not isinstance(
                inp.type,
                (
                    Int64TensorType,
                    StringTensorType,
                    Int32TensorType,
                    FloatTensorType,
                    DoubleTensorType,
                ),
            ):
                raise NotImplementedError(
                    "{} input datatype not yet supported. "
                    "You may raise an issue at "
                    "https://github.com/onnx/sklearn-onnx/issues"
                    "".format(type(inp.type))
                )

            if all_shapes[index_inputs] == 1:
                assert index_in_input == 0
                afeat = False
            else:
                afeat = True
            enum_cats.append((afeat, index_in_input, inp.full_name, cats, inp.type))
    else:
        inp = operator.inputs[0]
        enum_cats = [
            (True, i, inp.full_name, cats, inp.type)
            for i, cats in enumerate(ohe_op.categories_)
        ]

    result, categories_len = [], 0
    for index, enum_c in enumerate(enum_cats):
        afeat, index_in, name, categories, inp_type = enum_c
        container.debug(
            "[conv.OneHotEncoder] cat %r/%r name=%r type=%r",
            index + 1,
            len(enum_cats),
            name,
            inp_type,
        )
        if len(categories) == 0:
            continue
        if afeat:
            index_name = scope.get_unique_variable_name(name + str(index_in))
            container.add_initializer(
                index_name, onnx_proto.TensorProto.INT64, [1], [index_in]
            )
            out_name = scope.get_unique_variable_name(name + str(index_in))
            container.add_node(
                "Gather",
                [name, index_name],
                out_name,
                axis=1,
                name=scope.get_unique_operator_name("Gather"),
            )
            name = out_name

        attrs = {"name": scope.get_unique_operator_name("OneHotEncoder")}
        attrs["zeros"] = 1 if ohe_op.handle_unknown == "ignore" else 0

        if isinstance(inp_type, (Int64TensorType, Int32TensorType)):
            attrs["cats_int64s"] = categories.astype(np.int64)
        elif isinstance(inp_type, StringTensorType):
            attrs["cats_strings"] = np.array(
                [str(s).encode("utf-8") for s in categories]
            )
        elif isinstance(inp_type, (FloatTensorType, DoubleTensorType)):
            # The converter checks that categories can be casted into
            # integers. String is not allowed here.
            # Input type is casted into int64.
            for c in categories:
                try:
                    ci = int(c)
                except TypeError:
                    raise RuntimeError(
                        "Category '{}' cannot be casted into int.".format(c)
                    )
                if ci != c:
                    raise RuntimeError(
                        "Category %r is not an int64. "
                        "The converter only supports string and int64 "
                        "categories not %r." % (c, type(c))
                    )
            attrs["cats_int64s"] = categories.astype(np.int64)
        else:
            raise RuntimeError(
                "Input type {} is not supported for OneHotEncoder. "
                "Ideally, it should either be integer or strings.".format(inp_type)
            )

        ohe_output = scope.get_unique_variable_name(name + "out")

        if "cats_int64s" in attrs:
            # Let's cast this input in int64.
            cast_feature = scope.get_unique_variable_name(name + "cast")
            apply_cast(
                scope, name, cast_feature, container, to=onnx_proto.TensorProto.INT64
            )
            name = cast_feature

        container.add_node(
            "OneHotEncoder", name, ohe_output, op_domain="ai.onnx.ml", **attrs
        )
        if (
            hasattr(ohe_op, "drop_idx_")
            and ohe_op.drop_idx_ is not None
            and ohe_op.drop_idx_[index] is not None
        ):
            extracted_outputs_name = scope.get_unique_variable_name("extracted_outputs")
            indices_to_keep_name = scope.get_unique_variable_name("indices_to_keep")
            indices_to_keep = np.delete(
                np.arange(len(categories)), ohe_op.drop_idx_[index]
            )
            container.add_initializer(
                indices_to_keep_name,
                onnx_proto.TensorProto.INT64,
                indices_to_keep.shape,
                indices_to_keep,
            )
            container.add_node(
                "Gather",
                [ohe_output, indices_to_keep_name],
                extracted_outputs_name,
                axis=-1,
                name=scope.get_unique_operator_name("Gather"),
            )
            ohe_output, categories = extracted_outputs_name, indices_to_keep

        result.append(ohe_output)
        categories_len += len(categories)

    concat_result_name = scope.get_unique_variable_name("concat_result")
    apply_concat(scope, result, concat_result_name, container, axis=-1)

    reshape_input = concat_result_name
    if np.issubdtype(ohe_op.dtype, np.signedinteger):
        reshape_input = scope.get_unique_variable_name("cast")
        apply_cast(
            scope,
            concat_result_name,
            reshape_input,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    apply_reshape(
        scope,
        reshape_input,
        operator.output_full_names,
        container,
        desired_shape=(-1, categories_len),
    )


register_converter("SklearnOneHotEncoder", convert_sklearn_one_hot_encoder)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common.data_types import guess_proto_type
from ..common._apply_operation import (
    apply_add,
    apply_cast,
    apply_concat,
    apply_identity,
    apply_reshape,
    apply_sub,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def _forward_pass(scope, container, model, activations, proto_dtype):
    """
    Perform a forward pass on the network by computing the values of
    the neurons in the hidden layers and the output layer.
    """
    activations_map = {
        "identity": "Identity",
        "tanh": "Tanh",
        "logistic": "Sigmoid",
        "relu": "Relu",
        "softmax": "Softmax",
    }

    out_activation_result_name = scope.get_unique_variable_name(
        "out_activations_result"
    )

    # Iterate over the hidden layers
    for i in range(model.n_layers_ - 1):
        coefficient_name = scope.get_unique_variable_name("coefficient")
        intercepts_name = scope.get_unique_variable_name("intercepts")
        mul_result_name = scope.get_unique_variable_name("mul_result")
        add_result_name = scope.get_unique_variable_name("add_result")

        container.add_initializer(
            coefficient_name,
            proto_dtype,
            model.coefs_[i].shape,
            model.coefs_[i].ravel(),
        )
        container.add_initializer(
            intercepts_name,
            proto_dtype,
            [1, len(model.intercepts_[i])],
            model.intercepts_[i],
        )

        container.add_node(
            "MatMul",
            [activations[i], coefficient_name],
            mul_result_name,
            name=scope.get_unique_operator_name("MatMul"),
        )
        apply_add(
            scope,
            [mul_result_name, intercepts_name],
            add_result_name,
            container,
            broadcast=1,
        )

        # For the hidden layers
        if (i + 1) != (model.n_layers_ - 1):
            activations_result_name = scope.get_unique_variable_name("next_activations")

            container.add_node(
                activations_map[model.activation],
                add_result_name,
                activations_result_name,
                name=scope.get_unique_operator_name(activations_map[model.activation]),
            )
            activations.append(activations_result_name)

    # For the last layer
    container.add_node(
        activations_map[model.out_activation_],
        add_result_name,
        out_activation_result_name,
        name=scope.get_unique_operator_name(activations_map[model.activation]),
    )
    activations.append(out_activation_result_name)

    return activations


def _predict(scope, input_name, container, model, proto_dtype):
    """
    This function initialises the input layer, calls _forward_pass()
    and returns the final layer.
    """
    cast_input_name = scope.get_unique_variable_name("cast_input")

    apply_cast(scope, input_name, cast_input_name, container, to=proto_dtype)

    # forward propagate
    activations = _forward_pass(scope, container, model, [cast_input_name], proto_dtype)
    return activations[-1]


def convert_sklearn_mlp_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for MLPClassifier.
    This function calls _predict() which returns the probability scores
    of the positive class in case of binary labels and class
    probabilities in case of multi-class. It then calculates probability
    scores for the negative class in case of binary labels. It
    calculates the class labels and sets the output.
    """
    mlp_op = operator.raw_operator
    classes = mlp_op.classes_
    class_type = onnx_proto.TensorProto.STRING

    argmax_output_name = scope.get_unique_variable_name("argmax_output")
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        "array_feature_extractor_result"
    )

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    y_pred = _predict(
        scope, operator.inputs[0].full_name, container, mlp_op, proto_dtype
    )

    if (
        np.issubdtype(mlp_op.classes_.dtype, np.floating)
        or mlp_op.classes_.dtype == np.bool_
    ):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(mlp_op.classes_.dtype, np.integer):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])

    if len(classes) == 2:
        unity_name = scope.get_unique_variable_name("unity")
        negative_class_proba_name = scope.get_unique_variable_name(
            "negative_class_proba"
        )
        container.add_initializer(unity_name, proto_dtype, [], [1])
        apply_sub(
            scope,
            [unity_name, y_pred],
            negative_class_proba_name,
            container,
            broadcast=1,
        )
        apply_concat(
            scope,
            [negative_class_proba_name, y_pred],
            operator.outputs[1].full_name,
            container,
            axis=1,
        )
    else:
        apply_identity(scope, y_pred, operator.outputs[1].full_name, container)

    if mlp_op._label_binarizer.y_type_ == "multilabel-indicator":
        binariser_output_name = scope.get_unique_variable_name("binariser_output")

        container.add_node(
            "Binarizer",
            y_pred,
            binariser_output_name,
            threshold=0.5,
            op_domain="ai.onnx.ml",
        )
        apply_cast(
            scope,
            binariser_output_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        classes_name = scope.get_unique_variable_name("classes")
        container.add_initializer(classes_name, class_type, classes.shape, classes)

        container.add_node(
            "ArgMax",
            operator.outputs[1].full_name,
            argmax_output_name,
            axis=1,
            name=scope.get_unique_operator_name("ArgMax"),
        )
        container.add_node(
            "ArrayFeatureExtractor",
            [classes_name, argmax_output_name],
            array_feature_extractor_result_name,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        )

        if class_type == onnx_proto.TensorProto.INT32:
            reshaped_result_name = scope.get_unique_variable_name("reshaped_result")

            apply_reshape(
                scope,
                array_feature_extractor_result_name,
                reshaped_result_name,
                container,
                desired_shape=(-1,),
            )
            apply_cast(
                scope,
                reshaped_result_name,
                operator.outputs[0].full_name,
                container,
                to=onnx_proto.TensorProto.INT64,
            )
        else:
            apply_reshape(
                scope,
                array_feature_extractor_result_name,
                operator.outputs[0].full_name,
                container,
                desired_shape=(-1,),
            )


def convert_sklearn_mlp_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for MLPRegressor.
    This function calls _predict() which returns the scores.
    """
    mlp_op = operator.raw_operator

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    y_pred = _predict(
        scope, operator.inputs[0].full_name, container, mlp_op, proto_dtype=proto_dtype
    )
    apply_reshape(
        scope, y_pred, operator.output_full_names, container, desired_shape=(-1, 1)
    )


register_converter(
    "SklearnMLPClassifier",
    convert_sklearn_mlp_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "output_class_labels": [False, True],
        "nocl": [True, False],
    },
)
register_converter("SklearnMLPRegressor", convert_sklearn_mlp_regressor)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.shape_calculator import calculate_linear_classifier_output_shapes


register_shape_calculator(
    "SklearnOneVsOneClassifier", calculate_linear_classifier_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from onnx import TensorProto
from onnx.helper import make_tensor
from onnx.numpy_helper import from_array
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_feature_hasher(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    out = operator.outputs
    op = operator.raw_operator
    if op.input_type != "string":
        raise RuntimeError(
            f"The converter for FeatureHasher only supports "
            f"input_type='string' not {op.input_type!r}."
        )

    # If option separator is not None, the converter assumes the input
    # is one string column, each element is a list of strings concatenated
    # with this separator.
    options = container.get_options(op, dict(separator=None))
    separator = options.get("separator", None)

    if separator is not None:
        # Let's split the columns
        delimiter = scope.get_unique_variable_name("delimiter")
        container.add_initializer(
            delimiter, TensorProto.STRING, [], [separator.encode("utf-8")]
        )
        empty_string = scope.get_unique_variable_name("empty_string")
        container.add_initializer(
            empty_string, TensorProto.STRING, [], ["".encode("utf-8")]
        )
        skip_empty = scope.get_unique_variable_name("delimiter")
        container.add_initializer(skip_empty, TensorProto.BOOL, [], [False])
        flat_shape = scope.get_unique_variable_name("flat_shape")
        container.add_initializer(flat_shape, TensorProto.INT64, [1], [-1])
        zero = scope.get_unique_variable_name("zero")
        container.add_initializer(zero, TensorProto.INT64, [1], [0])
        one = scope.get_unique_variable_name("one")
        container.add_initializer(one, TensorProto.INT64, [1], [1])

        to_concat = []
        for i, col_to_split in enumerate(operator.inputs):
            reshaped = scope.get_unique_variable_name(f"reshaped{i}")
            container.add_node(
                "Reshape", [col_to_split.full_name, flat_shape], [reshaped]
            )
            out_indices = scope.get_unique_variable_name(f"out_indices{i}")
            out_text = scope.get_unique_variable_name(f"out_text{i}")
            out_shape = scope.get_unique_variable_name(f"out_shape{i}")
            if len(separator) <= 1:
                container.add_node(
                    "StringSplit",
                    [reshaped, delimiter, skip_empty],
                    [out_indices, out_text, out_shape],
                    op_domain="ai.onnx.contrib",
                    op_version=1,
                )
            else:
                raise RuntimeError(
                    f"Only one character separators are supported but delimiter is {separator!r}."
                )
            shape = scope.get_unique_variable_name(f"shape{i}")
            container.add_node("Shape", [col_to_split.full_name], [shape])

            emptyi = scope.get_unique_variable_name(f"emptyi{i}")
            container.add_node(
                "ConstantOfShape",
                [out_shape],
                [emptyi],
                value=from_array(np.array([0], dtype=np.int64)),
            )
            emptyb = scope.get_unique_variable_name(f"emptyb{i}")
            container.add_node("Cast", [emptyi], [emptyb], to=TensorProto.BOOL)
            emptys = scope.get_unique_variable_name(f"emptys{i}")
            container.add_node("Where", [emptyb, empty_string, empty_string], [emptys])
            flat_split = scope.get_unique_variable_name(f"flat_split{i}")
            container.add_node(
                "ScatterND", [emptys, out_indices, out_text], [flat_split]
            )
            # shape_1 = scope.get_unique_variable_name(f"shape_1{i}")
            # container.add_node("Concat", [shape, flat_shape], [shape_1], axis=0)

            split = scope.get_unique_variable_name(f"split{i}")
            to_concat.append(split)
            # container.add_node("Reshape", [flat_split, shape_1], [split])
            container.add_node("Identity", [flat_split], [split])
        if len(to_concat) == 1:
            input_hasher = to_concat[0]
        else:
            input_hasher = scope.get_unique_variable_name("concatenated")
            container.add_node("Concat", to_concat, [input_hasher], axis=1)
    elif len(operator.inputs) == 1:
        X = operator.inputs[0]
        input_hasher = X.full_name
    else:
        raise RuntimeError(
            f"Only one input is expected but received "
            f"{[i.name for i in operator.inputs]}."
        )

    hashed_ = scope.get_unique_variable_name("hashed_")
    container.add_node(
        "MurmurHash3",
        input_hasher,
        hashed_,
        positive=0,
        seed=0,
        op_domain="com.microsoft",
        op_version=1,
    )
    hashed = scope.get_unique_variable_name("hashed")
    container.add_node("Cast", hashed_, hashed, to=TensorProto.INT64)

    if op.dtype in (np.float32, np.float64, np.int64):
        cst_neg = -1
    else:
        cst_neg = 4294967295

    infinite = scope.get_unique_variable_name("infinite")
    container.add_initializer(infinite, TensorProto.INT64, [1], [-2147483648])

    infinite2 = scope.get_unique_variable_name("infinite2")
    container.add_initializer(infinite2, TensorProto.INT64, [1], [cst_neg])

    infinite_n = scope.get_unique_variable_name("infinite_n")
    container.add_initializer(
        infinite_n, TensorProto.INT64, [1], [2147483647 - (op.n_features - 1)]
    )

    zero = scope.get_unique_variable_name("zero")
    container.add_initializer(zero, TensorProto.INT64, [1], [0])

    one = scope.get_unique_variable_name("one")
    container.add_initializer(one, TensorProto.INT64, [1], [1])

    mone = scope.get_unique_variable_name("mone")
    container.add_initializer(mone, TensorProto.INT64, [1], [-1])

    mtwo = scope.get_unique_variable_name("mtwo")
    container.add_initializer(mtwo, TensorProto.INT64, [1], [-2])

    nf = scope.get_unique_variable_name("nf")
    container.add_initializer(nf, TensorProto.INT64, [1], [op.n_features])

    new_shape = scope.get_unique_variable_name("new_shape")
    container.add_initializer(new_shape, TensorProto.INT64, [2], [-1, 1])

    # values
    if op.alternate_sign:
        cmp = scope.get_unique_variable_name("cmp")
        container.add_node("GreaterOrEqual", [hashed, zero], cmp)
        values = scope.get_unique_variable_name("values")
        container.add_node("Where", [cmp, one, infinite2], values)
    else:
        mul = scope.get_unique_variable_name("mul")
        container.add_node("Mul", [hashed, zero], mul)
        values = scope.get_unique_variable_name("values")
        container.add_node("Add", [mul, one], values)

    values_reshaped = scope.get_unique_variable_name("values_reshaped")
    container.add_node("Reshape", [values, new_shape], values_reshaped)

    # indices
    cmp = scope.get_unique_variable_name("cmp_ind")
    container.add_node("Equal", [hashed, infinite], cmp)
    values_abs = scope.get_unique_variable_name("values_abs")
    container.add_node("Abs", hashed, values_abs)
    values_ind = scope.get_unique_variable_name("values_ind")
    container.add_node("Where", [cmp, infinite_n, values_abs], values_ind)
    indices = scope.get_unique_variable_name("indices")
    container.add_node("Mod", [values_ind, nf], indices)
    indices_reshaped = scope.get_unique_variable_name("indices_reshaped")
    container.add_node("Reshape", [indices, new_shape], indices_reshaped)

    # scatter
    zerot_ = scope.get_unique_variable_name("zerot_")
    container.add_node(
        "ConstantOfShape",
        [nf],
        zerot_,
        value=make_tensor("value", TensorProto.INT64, [1], [0]),
    )
    zerot = scope.get_unique_variable_name("zerot")
    container.add_node("Mul", [indices_reshaped, zerot_], zerot)

    final = scope.get_unique_variable_name("final")
    container.add_node(
        "ScatterElements", [zerot, indices_reshaped, values_reshaped], final, axis=1
    )

    # at this point, every string has been processed as if it were in
    # in a single columns.
    # in case there is more than one column, we need to reduce over
    # the last dimension
    input_shape = scope.get_unique_variable_name("input_shape")
    container.add_node("Shape", input_hasher, input_shape)
    shape_not_last = scope.get_unique_variable_name("shape_not_last")
    container.add_node("Slice", [input_shape, zero, mone], shape_not_last)
    final_shape = scope.get_unique_variable_name("final_last")
    container.add_node("Concat", [shape_not_last, mone, nf], final_shape, axis=0)
    final_reshaped = scope.get_unique_variable_name("final_reshaped")
    container.add_node("Reshape", [final, final_shape], final_reshaped)

    if op.dtype == np.float32:
        to = TensorProto.FLOAT
    elif op.dtype == np.float64:
        to = TensorProto.DOUBLE
    elif op.dtype in (np.int32, np.uint32, np.int64):
        to = TensorProto.INT64
    else:
        raise RuntimeError(
            f"Converter is not implemented for FeatureHasher.dtype={op.dtype}."
        )
    final_reshaped_cast = scope.get_unique_variable_name("final_reshaped_cast")
    container.add_node("Cast", [final_reshaped], final_reshaped_cast, to=to)
    container.add_node(
        "ReduceSum", [final_reshaped_cast, mtwo], out[0].full_name, keepdims=0
    )


register_converter(
    "SklearnFeatureHasher", convert_sklearn_feature_hasher, options={"separator": None}
)



# SPDX-License-Identifier: Apache-2.0

"""
Tools to help benchmarking.
"""
from timeit import Timer
import numpy


def measure_time(stmt, context, repeat=10, number=50, div_by_number=False):
    """
    Measures a statement and returns the results as a dictionary.

    :param stmt: string
    :param context: variable to know in a dictionary
    :param repeat: average over *repeat* experiment
    :param number: number of executions in one row
    :param div_by_number: divide by the number of executions
    :return: dictionary

    .. runpython::
        :showcode:

        from skl2onnx.tutorial import measure_time
        from math import cos

        res = measure_time("cos(x)", context=dict(cos=cos, x=5.))
        print(res)

    See `Timer.repeat <https://docs.python.org/3/library/
    timeit.html?timeit.Timer.repeat>`_
    for a better understanding of parameter *repeat* and *number*.
    The function returns a duration corresponding to
    *number* times the execution of the main statement.
    """
    tim = Timer(stmt, globals=context)
    res = numpy.array(tim.repeat(repeat=repeat, number=number))
    if div_by_number:
        res /= number
    mean = numpy.mean(res)
    dev = numpy.mean(res**2)
    dev = (dev - mean**2) ** 0.5
    mes = dict(
        average=mean,
        deviation=dev,
        min_exec=numpy.min(res),
        max_exec=numpy.max(res),
        repeat=repeat,
        number=number,
    )
    return mes



# SPDX-License-Identifier: Apache-2.0


import onnx
from onnx.helper import make_graph, make_model
from onnx.defs import get_all_schemas_with_history


def _build_op_version():
    res = {}
    for schema in get_all_schemas_with_history():
        dom = schema.domain
        name = schema.name
        vers = schema.since_version
        if (dom, name) not in res:
            res[dom, name] = set()
        res[dom, name].add(vers)
    op_versions = {}
    for k, v in res.items():
        op_versions[k] = list(sorted(v))
    return op_versions


def _check_possible_opset(versions, nodes, domain, old_opset, new_opset):
    if old_opset >= new_opset:
        raise RuntimeError("Condition new_opset > old_opset is not true.")
    for node in nodes:
        name = node.op_type
        vers = versions.get((domain, name), None)
        if vers is None:
            # custom operator
            continue
        for v in vers:
            if v > old_opset and v <= new_opset:
                raise RuntimeError(
                    "Operator '{}' (domain: '{}') was updated "
                    "in opset {} in ]{}, {}]."
                    "".format(name, domain, v, old_opset, new_opset)
                )


def upgrade_opset_number(model, new_opsets):
    """
    Upgrades the domain opsets. It checks if that's
    possible. It means ONNX specifications do not
    propose a new version of every operator between
    the current opset and the new one.

    :param model: *ONNX* model
    :param new_opdets: integer or dictionary
        ``{ domain: opset }``
    :return: modified model
    """
    if isinstance(new_opsets, int):
        new_opsets = {"": new_opsets}
    graph = make_graph(
        model.graph.node,
        model.graph.name,
        model.graph.input,
        model.graph.input,
        model.graph.initializer,
    )
    onnx_model = make_model(graph)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:
        values = {p.key: p.value for p in model.metadata_props}
        onnx.helper.set_model_props(onnx_model, values)

    # fix opset import
    versions = _build_op_version()
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()
        op_set.domain = oimp.domain
        if oimp.domain in new_opsets:
            _check_possible_opset(
                versions,
                model.graph.node,
                oimp.domain,
                oimp.version,
                new_opsets[oimp.domain],
            )
            op_set.version = new_opsets[oimp.domain]
        else:
            op_set.version = oimp.version
    return onnx_model



# SPDX-License-Identifier: Apache-2.0


import copy
from ..common.data_types import FloatTensorType, Int64TensorType, DoubleTensorType
from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_polynomial_features(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    model = operator.raw_operator
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = [N, model.n_output_features_]


register_shape_calculator(
    "SklearnPolynomialFeatures", calculate_sklearn_polynomial_features
)



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_array_feature_extractor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Extracts a subset of columns. This is used by *ColumnTransformer*.
    """
    column_indices_name = scope.get_unique_variable_name("column_indices")

    for i, ind in enumerate(operator.column_indices):
        if not isinstance(ind, int):
            raise RuntimeError(
                (
                    "Column {0}:'{1}' indices must be specified "
                    "as integers. This error may happen when "
                    "column names are used to define a "
                    "ColumnTransformer. Column name in input data "
                    "do not necessarily match input variables "
                    "defined for the ONNX model."
                ).format(i, ind)
            )
    container.add_initializer(
        column_indices_name,
        onnx_proto.TensorProto.INT64,
        [len(operator.column_indices)],
        operator.column_indices,
    )

    container.add_node(
        "ArrayFeatureExtractor",
        [operator.inputs[0].full_name, column_indices_name],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        op_domain="ai.onnx.ml",
    )


register_converter(
    "SklearnArrayFeatureExtractor", convert_sklearn_array_feature_extractor
)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType,
    Int64TensorType,
    DoubleTensorType,
    StringTensorType,
)
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_imputer_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C_1], ..., [N, C_n] ---> [N, C_1 + ... + C_n]

    It's possible to receive multiple inputs so we need to concatenate
    them along C-axis. The produced tensor's shape is used as the
    output shape.
    """
    check_input_and_output_numbers(
        operator, input_count_range=[1, None], output_count_range=1
    )
    check_input_and_output_types(
        operator,
        good_input_types=[
            FloatTensorType,
            Int64TensorType,
            DoubleTensorType,
            StringTensorType,
        ],
    )
    output = operator.outputs[0]
    for variable in operator.inputs:
        if not isinstance(variable.type, type(output.type)):  # noqa
            raise RuntimeError(
                "Inputs and outputs should have the same type "
                "%r != %r." % (type(variable.type), type(output.type))
            )

    N = operator.inputs[0].get_first_dimension()
    C = 0
    for variable in operator.inputs:
        if variable.type.shape[1] is not None:
            C += variable.type.shape[1]
        else:
            C = None
            break

    output.type.shape = [N, C]


register_shape_calculator("SklearnImputer", calculate_sklearn_imputer_output_shapes)
register_shape_calculator(
    "SklearnSimpleImputer", calculate_sklearn_imputer_output_shapes
)
register_shape_calculator("SklearnBinarizer", calculate_sklearn_imputer_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.base import is_regressor
from sklearn.svm import LinearSVC
from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_concat,
    apply_identity,
    apply_mul,
    apply_reshape,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_normalization
from ..common._apply_operation import (
    apply_slice,
    apply_sub,
    apply_cast,
    apply_abs,
    apply_add,
    apply_div,
)
from ..common.utils_classifier import _finalize_converter_classes
from ..common.data_types import guess_proto_type, Int64TensorType
from ..algebra.onnx_ops import OnnxReshape, OnnxShape, OnnxSlice, OnnxTile
from .._supported_operators import sklearn_operator_name_map


def _iteration_one_versus(
    scope,
    container,
    inputs,
    i,
    estimator,
    cl_type,
    proto_dtype,
    use_raw_scores=True,
    prob_shape=None,
):
    op_type = sklearn_operator_name_map[type(estimator)]

    this_operator = scope.declare_local_operator(op_type, raw_model=estimator)
    this_operator.inputs = inputs

    if is_regressor(estimator):
        score_name = scope.declare_local_variable("score_%d" % i, cl_type())
        this_operator.outputs.append(score_name)

        if hasattr(estimator, "coef_") and len(estimator.coef_.shape) == 2:
            raise RuntimeError(
                "OneVsRestClassifier or OneVsOneClassifier accepts "
                "regressor with only one target."
            )
        p1 = score_name.onnx_name
    else:
        if container.has_options(estimator, "raw_scores"):
            container.add_options(id(estimator), {"raw_scores": use_raw_scores})
            scope.add_options(id(estimator), {"raw_scores": use_raw_scores})
        label_name = scope.declare_local_variable("label_%d" % i, Int64TensorType())
        prob_name = scope.declare_local_variable("proba_%d" % i, cl_type())
        this_operator.outputs.append(label_name)
        this_operator.outputs.append(prob_name)

        # gets the probability for the class 1
        p1 = scope.get_unique_variable_name("probY_%d" % i)
        if isinstance(estimator, LinearSVC):
            apply_identity(scope, prob_name.onnx_name, p1, container)
        else:
            apply_slice(
                scope,
                prob_name.onnx_name,
                p1,
                container,
                starts=[1],
                ends=[2],
                axes=[1],
            )
    return None, None, p1


def convert_one_vs_rest_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts a *OneVsRestClassifier* into *ONNX* format.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))["nocl"]:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__
            )
        )
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]
    probs_names = []
    cl_type = operator.inputs[0].type.__class__
    prob_shape = None
    for i, estimator in enumerate(op.estimators_):
        prob_shape, _, p1 = _iteration_one_versus(
            scope,
            container,
            operator.inputs,
            i,
            estimator,
            cl_type,
            proto_dtype,
            use_raw_scores,
            prob_shape=prob_shape,
        )
        probs_names.append(p1)

    if op.multilabel_:
        # concatenates outputs
        conc_name = operator.outputs[1].full_name
        apply_concat(scope, probs_names, conc_name, container, axis=1)

        # builds the labels (matrix with integer)
        # scikit-learn may use probabilities or raw score
        # but ONNX converters only uses probabilities.
        # https://github.com/scikit-learn/scikit-learn/sklearn/
        # multiclass.py#L290
        # Raw score would mean: scores = conc_name.
        thresh_name = scope.get_unique_variable_name("thresh")
        container.add_initializer(
            thresh_name, proto_dtype, [1, len(op.classes_)], [0.5] * len(op.classes_)
        )
        scores = scope.get_unique_variable_name("threshed")
        apply_sub(scope, [conc_name, thresh_name], scores, container)

        # sign
        signed_input = scope.get_unique_variable_name("signed")
        container.add_node(
            "Sign",
            [scores],
            [signed_input],
            name=scope.get_unique_operator_name("Sign"),
        )
        # clip
        signed_input_cast = scope.get_unique_variable_name("signed_int64")
        apply_cast(
            scope,
            signed_input,
            signed_input_cast,
            container,
            to=onnx_proto.TensorProto.INT64,
        )

        label_name = scope.get_unique_variable_name("label")
        if container.target_opset <= 11:
            abs_name = scope.get_unique_variable_name("abs")
            add_name = scope.get_unique_variable_name("add")
            cst_2 = scope.get_unique_variable_name("cst2")
            container.add_initializer(cst_2, onnx_proto.TensorProto.INT64, [1], [2])
            apply_abs(scope, [signed_input_cast], [abs_name], container)
            apply_add(scope, [signed_input_cast, abs_name], [add_name], container)
            apply_div(scope, [add_name, cst_2], [label_name], container)
        else:
            zero_cst = scope.get_unique_variable_name("zero")
            container.add_initializer(zero_cst, onnx_proto.TensorProto.INT64, [], [0])
            container.add_node(
                "Clip",
                [signed_input_cast, zero_cst],
                [label_name],
                name=scope.get_unique_operator_name("Clip"),
            )
        apply_reshape(
            scope,
            [label_name],
            [operator.outputs[0].full_name],
            container,
            desired_shape=(-1, op.n_classes_),
        )
    else:
        # concatenates outputs
        conc_name = scope.get_unique_variable_name("concatenated")
        apply_concat(scope, probs_names, conc_name, container, axis=1)
        if len(op.estimators_) == 1:
            zeroth_col_name = scope.get_unique_variable_name("zeroth_col")
            merged_prob_name = scope.get_unique_variable_name("merged_prob")
            unit_float_tensor_name = scope.get_unique_variable_name("unit_float_tensor")
            if use_raw_scores:
                container.add_initializer(
                    unit_float_tensor_name, proto_dtype, [], [-1.0]
                )
                apply_mul(
                    scope,
                    [unit_float_tensor_name, conc_name],
                    zeroth_col_name,
                    container,
                    broadcast=1,
                )
            else:
                container.add_initializer(
                    unit_float_tensor_name, proto_dtype, [], [1.0]
                )
                apply_sub(
                    scope,
                    [unit_float_tensor_name, conc_name],
                    zeroth_col_name,
                    container,
                    broadcast=1,
                )
            apply_concat(
                scope, [zeroth_col_name, conc_name], merged_prob_name, container, axis=1
            )
            conc_name = merged_prob_name

        if use_raw_scores:
            apply_identity(scope, conc_name, operator.outputs[1].full_name, container)
        else:
            # normalizes the outputs
            apply_normalization(
                scope, conc_name, operator.outputs[1].full_name, container, axis=1, p=1
            )

        # extracts the labels
        label_name = scope.get_unique_variable_name("label_name")
        container.add_node(
            "ArgMax",
            conc_name,
            label_name,
            name=scope.get_unique_operator_name("ArgMax"),
            axis=1,
        )

        _finalize_converter_classes(
            scope,
            label_name,
            operator.outputs[0].full_name,
            container,
            op.classes_,
            proto_dtype,
        )


def convert_constant_predictor_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts a *_ConstantPredictor* into *ONNX* format.
    """
    op_version = container.target_opset
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    dtype = {
        onnx_proto.TensorProto.DOUBLE: np.float64,
        onnx_proto.TensorProto.FLOAT: np.float32,
    }
    shape = OnnxShape(operator.inputs[0].full_name, op_version=op_version)
    first = OnnxSlice(
        shape,
        np.array([0], dtype=np.int64),
        np.array([1], dtype=np.int64),
        op_version=op_version,
    )
    y = op.y_.astype(dtype[proto_dtype]).ravel()
    labels = OnnxTile(
        y.astype(np.int64),
        first,
        op_version=op_version,
        output_names=[operator.outputs[0].full_name],
    )

    cst = np.hstack([(1 - y).astype(y.dtype), y])
    proba_flat = OnnxTile(cst, first, op_version=op_version)
    proba_reshape = OnnxReshape(
        proba_flat,
        np.array([-1, 2], dtype=np.int64),
        output_names=[operator.outputs[1].full_name],
        op_version=op_version,
    )

    labels.add_to(scope, container)
    proba_reshape.add_to(scope, container)


register_converter(
    "SklearnOneVsRestClassifier",
    convert_one_vs_rest_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)

register_converter(
    "Sklearn_ConstantPredictor",
    convert_constant_predictor_classifier,
    options={"zipmap": [True, False, "columns"], "nocl": [True, False]},
)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator


def pipeline_shape_calculator(operator):
    pass


def feature_union_shape_calculator(operator):
    pass


def column_transformer_shape_calculator(operator):
    pass


register_shape_calculator("SklearnPipeline", pipeline_shape_calculator)
register_shape_calculator("SklearnFeatureUnion", feature_union_shape_calculator)
register_shape_calculator(
    "SklearnColumnTransformer", column_transformer_shape_calculator
)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType
from ..common.shape_calculator import calculate_linear_classifier_output_shapes


def calculate_constant_predictor_output_shapes(operator):
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = Int64TensorType([N])
    operator.outputs[1].type.shape = [N, 2]


register_shape_calculator(
    "Sklearn_ConstantPredictor", calculate_constant_predictor_output_shapes
)

register_shape_calculator(
    "SklearnOneVsRestClassifier", calculate_linear_classifier_output_shapes
)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_normalizer
from ..common.data_types import DoubleTensorType
from .common import concatenate_variables


def convert_sklearn_normalizer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    if len(operator.inputs) > 1:
        # If there are multiple input tensors,
        # we combine them using a FeatureVectorizer
        feature_name = concatenate_variables(scope, operator.inputs, container)
    else:
        # No concatenation is needed, we just use the first variable's name
        feature_name = operator.inputs[0].full_name
    op = operator.raw_operator
    norm_map = {"max": "MAX", "l1": "L1", "l2": "L2"}
    if op.norm in norm_map:
        norm = norm_map[op.norm]
    else:
        raise RuntimeError(
            "Invalid norm '%s'. You may raise an issue"
            "at https://github.com/onnx/sklearn-onnx/"
            "issues." % op.norm
        )
    use_float = type(operator.inputs[0].type) not in (DoubleTensorType,)
    apply_normalizer(
        scope,
        feature_name,
        operator.outputs[0].full_name,
        container,
        norm=norm,
        use_float=use_float,
    )


register_converter("SklearnNormalizer", convert_sklearn_normalizer)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._apply_operation import (
    apply_add,
    apply_cast,
    apply_clip,
    apply_concat,
    apply_div,
    apply_exp,
    apply_identity,
    apply_mul,
    apply_reciprocal,
    apply_reshape,
    apply_sub,
)
from ..common.data_types import (
    BooleanTensorType,
    Int64TensorType,
    guess_numpy_type,
    guess_proto_type,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto


def _decision_function(scope, operator, container, model, proto_type):
    """Predict for linear model.
    score = X * coefficient + intercept
    """
    coef_name = scope.get_unique_variable_name("coef")
    intercept_name = scope.get_unique_variable_name("intercept")
    matmul_result_name = scope.get_unique_variable_name("matmul_result")
    score_name = scope.get_unique_variable_name("score")
    coef = model.coef_.T

    container.add_initializer(coef_name, proto_type, coef.shape, coef.ravel())
    container.add_initializer(
        intercept_name, proto_type, model.intercept_.shape, model.intercept_
    )

    input_name = operator.inputs[0].full_name
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope, operator.input_full_names, cast_input_name, container, to=proto_type
        )
        input_name = cast_input_name

    container.add_node(
        "MatMul",
        [input_name, coef_name],
        matmul_result_name,
        name=scope.get_unique_operator_name("MatMul"),
    )
    apply_add(
        scope, [matmul_result_name, intercept_name], score_name, container, broadcast=0
    )
    return score_name


def _handle_zeros(
    scope, container, scores, proba, reduced_proba, num_classes, proto_type
):
    """Handle cases where reduced_proba values are zeros to avoid NaNs in
    class probability scores because of divide by 0 when we calculate
    proba / reduced_proba in _normalise_proba().
    This is done by replacing reduced_proba values of 0s with
    num_classes and corresponding proba values with 1.
    """
    num_classes_name = scope.get_unique_variable_name("num_classes")
    bool_reduced_proba_name = scope.get_unique_variable_name("bool_reduced_proba")
    bool_not_reduced_proba_name = scope.get_unique_variable_name(
        "bool_not_reduced_proba"
    )
    not_reduced_proba_name = scope.get_unique_variable_name("not_reduced_proba")
    proba_updated_name = scope.get_unique_variable_name("proba_updated")
    mask_name = scope.get_unique_variable_name("mask")
    reduced_proba_updated_name = scope.get_unique_variable_name("reduced_proba_updated")

    container.add_initializer(num_classes_name, proto_type, [], [num_classes])

    apply_cast(
        scope,
        reduced_proba,
        bool_reduced_proba_name,
        container,
        to=onnx_proto.TensorProto.BOOL,
    )
    container.add_node(
        "Not",
        bool_reduced_proba_name,
        bool_not_reduced_proba_name,
        name=scope.get_unique_operator_name("Not"),
    )
    apply_cast(
        scope,
        bool_not_reduced_proba_name,
        not_reduced_proba_name,
        container,
        to=proto_type,
    )
    apply_add(
        scope,
        [proba, not_reduced_proba_name],
        proba_updated_name,
        container,
        broadcast=1,
    )
    apply_mul(
        scope,
        [not_reduced_proba_name, num_classes_name],
        mask_name,
        container,
        broadcast=1,
    )
    apply_add(
        scope,
        [reduced_proba, mask_name],
        reduced_proba_updated_name,
        container,
        broadcast=0,
    )
    return proba_updated_name, reduced_proba_updated_name


def _normalise_proba(
    scope, operator, container, scores, proba, num_classes, unity_name, proto_type
):
    reduced_proba_name = scope.get_unique_variable_name("reduced_proba")
    sub_result_name = scope.get_unique_variable_name("sub_result")

    if num_classes == 2:
        apply_sub(scope, [unity_name, proba], sub_result_name, container, broadcast=1)
        apply_concat(
            scope,
            [sub_result_name, proba],
            operator.outputs[1].full_name,
            container,
            axis=1,
        )
    else:
        if container.target_opset < 13:
            container.add_node(
                "ReduceSum",
                proba,
                reduced_proba_name,
                axes=[1],
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        else:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                "ReduceSum",
                [proba, axis_name],
                reduced_proba_name,
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        proba_updated, reduced_proba_updated = _handle_zeros(
            scope, container, scores, proba, reduced_proba_name, num_classes, proto_type
        )
        apply_div(
            scope,
            [proba_updated, reduced_proba_updated],
            operator.outputs[1].full_name,
            container,
            broadcast=1,
        )
    return operator.outputs[1].full_name


def _predict_proba_log(scope, operator, container, scores, num_classes, proto_type):
    """Probability estimation for SGDClassifier with loss=log (or log_loss)
    and Logistic Regression.
    Positive class probabilities are computed as
        1. / (1. + exp(-scores))
        multiclass is handled by normalising that over all classes.
    """
    if num_classes >= 3 or container.target_opset < 13:
        negated_scores_name = scope.get_unique_variable_name("negated_scores")
        negate_name = scope.get_unique_variable_name("negate")
        exp_result_name = scope.get_unique_variable_name("exp_result")
        unity_name = scope.get_unique_variable_name("unity")
        add_result_name = scope.get_unique_variable_name("add_result")
        proba_name = scope.get_unique_variable_name("proba")

        container.add_initializer(negate_name, proto_type, [], [-1])
        container.add_initializer(unity_name, proto_type, [], [1])

        apply_mul(
            scope, [scores, negate_name], negated_scores_name, container, broadcast=1
        )
        apply_exp(scope, negated_scores_name, exp_result_name, container)
        apply_add(
            scope,
            [exp_result_name, unity_name],
            add_result_name,
            container,
            broadcast=1,
        )
        apply_reciprocal(scope, add_result_name, proba_name, container)
        return _normalise_proba(
            scope,
            operator,
            container,
            scores,
            proba_name,
            num_classes,
            unity_name,
            proto_type,
        )

    # Sigmoid cannot be used for num_classes > 2 because
    # onnxruntime has a different implementation than numpy.
    # It introduces discrepancies when x < 1e16.
    # Below that threshold, Sigmoid must be replaced by Exp
    # because Sigmoid is not an increasing function.
    sigmo = scope.get_unique_variable_name("sigmoid")
    container.add_node(
        "Sigmoid", [scores], [sigmo], name=scope.get_unique_operator_name("Sigmoid")
    )

    unity_name = scope.get_unique_variable_name("unity")
    container.add_initializer(unity_name, proto_type, [1], [1])

    sigmo_0 = scope.get_unique_variable_name("sigmo_0")
    container.add_node(
        "Sub",
        [unity_name, sigmo],
        [sigmo_0],
        name=scope.get_unique_operator_name("Sub"),
    )
    apply_concat(
        scope, [sigmo_0, sigmo], [operator.outputs[1].full_name], container, axis=1
    )
    return operator.outputs[1].full_name


def _predict_proba_modified_huber(
    scope, operator, container, scores, num_classes, proto_type
):
    """Probability estimation for SGDClassifier with
    loss=modified_huber.
    Multiclass probability estimates are derived from binary
    estimates by normalisation.
    Binary probability estimates are given by
    (clip(scores, -1, 1) + 1) / 2.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    unity_name = scope.get_unique_variable_name("unity")
    constant_name = scope.get_unique_variable_name("constant")
    add_result_name = scope.get_unique_variable_name("add_result")
    proba_name = scope.get_unique_variable_name("proba")
    clipped_scores_name = scope.get_unique_variable_name("clipped_scores")

    container.add_initializer(unity_name, proto_type, [], [1])
    container.add_initializer(constant_name, proto_type, [], [2])

    apply_clip(
        scope,
        scores,
        clipped_scores_name,
        container,
        max=np.array(1, dtype=dtype),
        min=np.array(-1, dtype=dtype),
    )
    apply_add(
        scope,
        [clipped_scores_name, unity_name],
        add_result_name,
        container,
        broadcast=1,
    )
    apply_div(
        scope, [add_result_name, constant_name], proba_name, container, broadcast=1
    )
    return _normalise_proba(
        scope,
        operator,
        container,
        scores,
        proba_name,
        num_classes,
        unity_name,
        proto_type,
    )


def convert_sklearn_sgd_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """Converter for SGDClassifier."""
    sgd_op = operator.raw_operator
    classes = get_label_classes(scope, sgd_op)
    class_type = onnx_proto.TensorProto.STRING
    proto_type = guess_proto_type(operator.inputs[0].type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT

    if np.issubdtype(classes.dtype, np.floating) or classes.dtype == np.bool_:
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])

    classes_name = scope.get_unique_variable_name("classes")
    predicted_label_name = scope.get_unique_variable_name("predicted_label")
    final_label_name = scope.get_unique_variable_name("final_label")

    container.add_initializer(classes_name, class_type, classes.shape, classes)

    scores = _decision_function(scope, operator, container, sgd_op, proto_type)
    options = container.get_options(sgd_op, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]
    if sgd_op.loss in ("log", "log_loss") and not use_raw_scores:
        proba = _predict_proba_log(
            scope, operator, container, scores, len(classes), proto_type
        )
    elif sgd_op.loss == "modified_huber" and not use_raw_scores:
        proba = _predict_proba_modified_huber(
            scope, operator, container, scores, len(classes), proto_type
        )
    else:
        if len(classes) == 2:
            negate_name = scope.get_unique_variable_name("negate")
            negated_scores_name = scope.get_unique_variable_name("negated_scores")

            container.add_initializer(negate_name, proto_type, [], [-1])

            apply_mul(
                scope,
                [scores, negate_name],
                negated_scores_name,
                container,
                broadcast=1,
            )
            apply_concat(
                scope,
                [negated_scores_name, scores],
                operator.outputs[1].full_name,
                container,
                axis=1,
            )
        else:
            apply_identity(scope, scores, operator.outputs[1].full_name, container)
        proba = operator.outputs[1].full_name

    container.add_node(
        "ArgMax",
        proba,
        predicted_label_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
        keepdims=1,
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [classes_name, predicted_label_name],
        final_label_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )
    if class_type == onnx_proto.TensorProto.INT32:
        reshaped_final_label_name = scope.get_unique_variable_name(
            "reshaped_final_label"
        )

        apply_reshape(
            scope,
            final_label_name,
            reshaped_final_label_name,
            container,
            desired_shape=(-1,),
        )
        apply_cast(
            scope,
            reshaped_final_label_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        apply_reshape(
            scope,
            final_label_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=(-1,),
        )


register_converter(
    "SklearnSGDClassifier",
    convert_sklearn_sgd_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_cast
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from .._supported_operators import sklearn_operator_name_map


def convert_sklearn_cast(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    inp = operator.inputs[0]
    exptype = operator.outputs[0]
    res = exptype.type.to_onnx_type()
    et = res.tensor_type.elem_type
    apply_cast(scope, inp.full_name, exptype.full_name, container, to=et)


def convert_sklearn_cast_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    estimator = op.estimator

    op_type = sklearn_operator_name_map[type(estimator)]
    this_operator = scope.declare_local_operator(op_type, estimator)
    this_operator.inputs = operator.inputs

    cls = operator.inputs[0].type.__class__
    var_name = scope.declare_local_variable("cast_est", cls())
    this_operator.outputs.append(var_name)
    var_name = var_name.onnx_name

    exptype = operator.outputs[0]
    res = exptype.type.to_onnx_type()
    et = res.tensor_type.elem_type
    apply_cast(scope, var_name, exptype.full_name, container, to=et)


register_converter("SklearnCastTransformer", convert_sklearn_cast)
register_converter("SklearnCastRegressor", convert_sklearn_cast_regressor)
register_converter("SklearnCast", convert_sklearn_cast)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common._apply_operation import (
    apply_cast,
    apply_add,
    apply_sqrt,
    apply_div,
    apply_sub,
    apply_reshape,
)
from ..common.data_types import (
    BooleanTensorType,
    Int64TensorType,
    DoubleTensorType,
    guess_numpy_type,
    guess_proto_type,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto
from ..algebra.onnx_ops import (
    OnnxAdd,
    OnnxCast,
    OnnxExp,
    OnnxIdentity,
    OnnxMatMul,
    OnnxReshape,
    OnnxSigmoid,
)


def convert_sklearn_linear_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    use_linear_op = container.is_allowed({"LinearRegressor"})

    if not use_linear_op or type(operator.inputs[0].type) in (DoubleTensorType,):
        proto_dtype = guess_proto_type(operator.inputs[0].type)
        coef = scope.get_unique_variable_name("coef")
        if len(op.coef_.shape) == 1:
            model_coef = op.coef_.reshape((-1, 1))
        else:
            model_coef = op.coef_.T
        container.add_initializer(
            coef, proto_dtype, model_coef.shape, model_coef.ravel().tolist()
        )
        intercept = scope.get_unique_variable_name("intercept")
        value_intercept = op.intercept_.reshape((-1,))
        container.add_initializer(
            intercept,
            proto_dtype,
            value_intercept.shape,
            value_intercept.ravel().tolist(),
        )
        multiplied = scope.get_unique_variable_name("multiplied")
        container.add_node(
            "MatMul",
            [operator.inputs[0].full_name, coef],
            multiplied,
            name=scope.get_unique_operator_name("MatMul"),
        )
        resh = scope.get_unique_variable_name("resh")
        apply_add(scope, [multiplied, intercept], resh, container)
        last_dim = 1 if len(model_coef.shape) == 1 else model_coef.shape[-1]
        apply_reshape(
            scope,
            resh,
            operator.outputs[0].full_name,
            container,
            desired_shape=(-1, last_dim),
        )
        return

    op_type = "LinearRegressor"
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype not in (np.float32, np.float64):
        dtype = np.float32
    attrs = {"name": scope.get_unique_operator_name(op_type)}
    attrs["coefficients"] = op.coef_.astype(dtype).ravel()
    attrs["intercepts"] = np.array([op.intercept_], dtype=dtype).ravel()
    if len(op.coef_.shape) == 2:
        attrs["targets"] = op.coef_.shape[0]

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope,
            operator.input_full_names,
            cast_input_name,
            container,
            to=(
                onnx_proto.TensorProto.DOUBLE
                if dtype == np.float64
                else onnx_proto.TensorProto.FLOAT
            ),
        )
        input_name = cast_input_name
    container.add_node(
        op_type,
        input_name,
        operator.outputs[0].full_name,
        op_domain="ai.onnx.ml",
        **attrs,
    )


def convert_sklearn_bayesian_ridge(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    convert_sklearn_linear_regressor(scope, operator, container)

    op = operator.raw_operator
    options = container.get_options(op, dict(return_std=False))
    return_std = options["return_std"]
    if not return_std:
        return

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if hasattr(op, "normalize") and op.normalize:
        # if self.normalize:
        #     X = (X - self.X_offset_) / self.X_scale_
        offset = scope.get_unique_variable_name("offset")
        container.add_initializer(
            offset, proto_dtype, op.X_offset_.shape, op.X_offset_.ravel().tolist()
        )
        scale = scope.get_unique_variable_name("scale")
        container.add_initializer(
            scale, proto_dtype, op.X_scale_.shape, op.X_scale_.ravel().tolist()
        )
        centered = scope.get_unique_variable_name("centered")
        apply_sub(scope, [operator.inputs[0].full_name, offset], centered, container)
        scaled = scope.get_unique_variable_name("scaled")
        apply_div(scope, [centered, scale], scaled, container)
        input_name = scaled
    else:
        input_name = operator.inputs[0].full_name

    # sigmas_squared_data = (np.dot(X, self.sigma_) * X).sum(axis=1)
    sigma = scope.get_unique_variable_name("sigma")
    container.add_initializer(
        sigma, proto_dtype, op.sigma_.shape, op.sigma_.ravel().tolist()
    )
    sigmaed0 = scope.get_unique_variable_name("sigma0")
    container.add_node(
        "MatMul",
        [input_name, sigma],
        sigmaed0,
        name=scope.get_unique_operator_name("MatMul"),
    )
    sigmaed = scope.get_unique_variable_name("sigma")
    if container.target_opset < 13:
        container.add_node(
            "ReduceSum",
            sigmaed0,
            sigmaed,
            axes=[1],
            name=scope.get_unique_operator_name("ReduceSum"),
        )
    else:
        axis_name = scope.get_unique_variable_name("axis")
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            "ReduceSum",
            [sigmaed0, axis_name],
            sigmaed,
            name=scope.get_unique_operator_name("ReduceSum"),
        )

    # y_std = np.sqrt(sigmas_squared_data + (1. / self.alpha_))
    # return y_mean, y_std
    std0 = scope.get_unique_variable_name("std0")
    alphainv = scope.get_unique_variable_name("alphainv")
    container.add_initializer(alphainv, proto_dtype, [1], [1 / op.alpha_])
    apply_add(scope, [sigmaed, alphainv], std0, container)
    apply_sqrt(scope, std0, operator.outputs[1].full_name, container)


def convert_sklearn_poisson_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32

    if isinstance(X.type, Int64TensorType):
        input_var = OnnxCast(X, to=np.float32, op_version=opv)
    else:
        input_var = X

    intercept = (
        op.intercept_.astype(dtype)
        if len(op.intercept_.shape) > 0
        else np.array([op.intercept_], dtype=dtype)
    )
    eta = OnnxAdd(
        OnnxMatMul(input_var, op.coef_.astype(dtype), op_version=opv),
        intercept,
        op_version=opv,
    )

    if hasattr(op, "_link_instance"):
        # scikit-learn < 1.1
        from sklearn.linear_model._glm.link import IdentityLink, LogLink, LogitLink

        if isinstance(op._link_instance, IdentityLink):
            Y = OnnxIdentity(eta, op_version=opv)
        elif isinstance(op._link_instance, LogLink):
            Y = OnnxExp(eta, op_version=opv)
        elif isinstance(op._link_instance, LogitLink):
            Y = OnnxSigmoid(eta, op_version=opv)
        else:
            raise RuntimeError(
                "Unexpected type %r for _link_instance "
                "in operator type %r." % (type(op._link_instance), type(op))
            )
    else:
        # scikit-learn >= 1.1
        from sklearn._loss.loss import (
            AbsoluteError,
            HalfBinomialLoss,
            HalfGammaLoss,
            HalfPoissonLoss,
            HalfSquaredError,
            HalfTweedieLoss,
            HalfTweedieLossIdentity,
            PinballLoss,
        )

        loss = op._get_loss()
        if isinstance(
            loss,
            (AbsoluteError, HalfSquaredError, HalfTweedieLossIdentity, PinballLoss),
        ):
            Y = OnnxIdentity(eta, op_version=opv)
        elif isinstance(loss, (HalfPoissonLoss, HalfGammaLoss, HalfTweedieLoss)):
            Y = OnnxExp(eta, op_version=opv)
        elif isinstance(loss, HalfBinomialLoss):
            Y = OnnxSigmoid(eta, op_version=opv)
        else:
            raise RuntimeError(
                f"Unexpected type of link for {loss!r} loss " "in operator type {op!r}."
            )

    last_dim = 1 if len(op.coef_.shape) == 1 else op.coef_.shape[-1]
    final = OnnxReshape(
        Y,
        np.array([-1, last_dim], dtype=np.int64),
        op_version=opv,
        output_names=out[:1],
    )
    final.add_to(scope, container)


register_converter("SklearnLinearRegressor", convert_sklearn_linear_regressor)
register_converter("SklearnLinearSVR", convert_sklearn_linear_regressor)
register_converter(
    "SklearnBayesianRidge",
    convert_sklearn_bayesian_ridge,
    options={"return_std": [True, False]},
)
register_converter("SklearnPoissonRegressor", convert_sklearn_poisson_regressor)
register_converter("SklearnTweedieRegressor", convert_sklearn_poisson_regressor)



# SPDX-License-Identifier: Apache-2.0

from .onnx_operator import OnnxOperator


class OnnxCDist(OnnxOperator):
    """
    Defines a custom operator not defined by ONNX
    specifications but in onnxruntime.
    """

    since_version = 1
    expected_inputs = [("X", "T"), ("Y", "T")]
    expected_outputs = [("dist", "T")]
    input_range = [2, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = "com.microsoft"
    operator_name = "CDist"
    past_version = {}

    def __init__(self, X, Y, metric="sqeuclidean", op_version=None, **kwargs):
        """
        :param X: array or OnnxOperatorMixin
        :param Y: array or OnnxOperatorMixin
        :param metric: distance type
        :param dtype: *np.float32* or *np.float64*
        :param op_version: opset version
        :param kwargs: addition parameter
        """
        OnnxOperator.__init__(
            self, X, Y, metric=metric, op_version=op_version, **kwargs
        )


class OnnxSolve(OnnxOperator):
    """
    Defines a custom operator not defined by ONNX
    specifications but in onnxruntime.
    """

    since_version = 1
    expected_inputs = [("A", "T"), ("Y", "T")]
    expected_outputs = [("X", "T")]
    input_range = [2, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = "com.microsoft"
    operator_name = "Solve"
    past_version = {}

    def __init__(self, A, Y, lower=False, transposed=False, op_version=None, **kwargs):
        """
        :param A: array or OnnxOperatorMixin
        :param Y: array or OnnxOperatorMixin
        :param lower: see :epkg:`solve`
        :param transposed: see :epkg:`solve`
        :param dtype: *np.float32* or *np.float64*
        :param op_version: opset version
        :param kwargs: additional parameters
        """
        OnnxOperator.__init__(
            self,
            A,
            Y,
            lower=lower,
            transposed=transposed,
            op_version=op_version,
            **kwargs
        )



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.base import RegressorMixin, BaseEstimator

try:
    from sklearn.utils.validation import _deprecate_positional_args
except ImportError:

    def _deprecate_positional_args(x):
        return x  # noqa


class CastRegressor(RegressorMixin, BaseEstimator):  # noqa

    """
    Cast predictions into a specific types.
    This should be used to minimize the conversion
    of a pipeline using float32 instead of double
    when onnx do not support double.

    Parameters
    ----------
    estimator : regressor
        wrapped regressor
    dtype : numpy type,
        output are cast into that type
    """  # noqa

    @_deprecate_positional_args
    def __init__(self, estimator, *, dtype=np.float32):
        self.dtype = dtype
        self.estimator = estimator

    def _cast(self, a, name):
        try:
            a2 = a.astype(self.dtype)
        except ValueError:
            raise ValueError(
                "Unable to cast {} from {} into {}.".format(name, a.dtype, self.dtype)
            )
        return a2

    def fit(self, X, y=None, sample_weight=None):
        """
        Does nothing except checking *dtype* may be applied.
        """
        self.estimator.fit(X, y=y, sample_weight=sample_weight)
        return self

    def predict(self, X, y=None):
        """
        Predicts and casts the prediction.
        """
        return self._cast(self.estimator.predict(X), "predict(X)")

    def decision_function(self, X, y=None):
        """
        Calls *decision_function* and casts the outputs.
        """
        if not hasattr(self.estimator, "decision_function"):
            raise AttributeError(
                "%r object has no attribute 'decision_function'."
                % self.estimator.__class__.__name__
            )
        return self._cast(self.estimator.decision_function(X), "decision_function(X)")



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_mul
from ..common.data_types import guess_proto_type
from .._supported_operators import sklearn_operator_name_map


def convert_voting_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts a *VotingRegressor* into *ONNX* format.
    """
    op = operator.raw_operator
    proto_dtype = guess_proto_type(operator.outputs[0].type)

    vars_names = []
    for i, estimator in enumerate(op.estimators_):
        if estimator is None:
            continue

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = operator.inputs

        var_name = scope.declare_local_variable(
            "var_%d" % i, operator.outputs[0].type.__class__()
        )
        this_operator.outputs.append(var_name)
        var_name = var_name.onnx_name

        if op.weights is not None:
            val = op.weights[i] / op.weights.sum()
        else:
            val = 1.0 / len(op.estimators_)

        weights_name = scope.get_unique_variable_name("w%d" % i)
        container.add_initializer(weights_name, proto_dtype, [1], [val])
        wvar_name = scope.get_unique_variable_name("wvar_%d" % i)
        apply_mul(scope, [var_name, weights_name], wvar_name, container, broadcast=1)

        flat_name = scope.get_unique_variable_name("fvar_%d" % i)
        container.add_node("Flatten", wvar_name, flat_name)
        vars_names.append(flat_name)

    container.add_node(
        "Sum",
        vars_names,
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name("Sum"),
    )


register_converter("SklearnVotingRegressor", convert_voting_regressor)



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common._apply_operation import apply_cast, apply_div, apply_sqrt, apply_sub
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    Int64TensorType,
    DoubleTensorType,
    FloatTensorType,
    guess_proto_type,
)


def convert_truncated_svd(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    # Create alias for the scikit-learn truncated SVD model we
    # are going to convert
    svd = operator.raw_operator
    if isinstance(operator.inputs[0].type, DoubleTensorType):
        proto_dtype = guess_proto_type(operator.inputs[0].type)
    else:
        proto_dtype = guess_proto_type(FloatTensorType())
    # Transpose [K, C] matrix to [C, K], where C/K is the
    # input/transformed feature dimension
    transform_matrix = svd.components_.transpose()
    transform_matrix_name = scope.get_unique_variable_name("transform_matrix")
    # Put the transformation into an ONNX tensor
    container.add_initializer(
        transform_matrix_name,
        proto_dtype,
        transform_matrix.shape,
        transform_matrix.flatten(),
    )

    input_name = operator.inputs[0].full_name
    if isinstance(operator.inputs[0].type, Int64TensorType):
        cast_output_name = scope.get_unique_variable_name("cast_output")

        apply_cast(
            scope,
            input_name,
            cast_output_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        input_name = cast_output_name

    if operator.type == "SklearnTruncatedSVD":
        # Create the major operator, a matrix multiplication.
        container.add_node(
            "MatMul",
            [input_name, transform_matrix_name],
            operator.outputs[0].full_name,
            name=operator.full_name,
        )
    else:  # PCA
        if svd.mean_ is not None:
            mean_name = scope.get_unique_variable_name("mean")
            sub_result_name = scope.get_unique_variable_name("sub_result")

            container.add_initializer(
                mean_name, proto_dtype, svd.mean_.shape, svd.mean_
            )

            # Subtract mean from input tensor
            apply_sub(
                scope, [input_name, mean_name], sub_result_name, container, broadcast=1
            )
        else:
            sub_result_name = input_name
        if svd.whiten:
            explained_variance_name = scope.get_unique_variable_name(
                "explained_variance"
            )
            explained_variance_root_name = scope.get_unique_variable_name(
                "explained_variance_root"
            )
            matmul_result_name = scope.get_unique_variable_name("matmul_result")

            container.add_initializer(
                explained_variance_name,
                proto_dtype,
                svd.explained_variance_.shape,
                svd.explained_variance_,
            )

            container.add_node(
                "MatMul",
                [sub_result_name, transform_matrix_name],
                matmul_result_name,
                name=scope.get_unique_operator_name("MatMul"),
            )
            apply_sqrt(
                scope, explained_variance_name, explained_variance_root_name, container
            )
            apply_div(
                scope,
                [matmul_result_name, explained_variance_root_name],
                operator.outputs[0].full_name,
                container,
                broadcast=1,
            )
        else:
            container.add_node(
                "MatMul",
                [sub_result_name, transform_matrix_name],
                operator.outputs[0].full_name,
                name=scope.get_unique_operator_name("MatMul"),
            )


register_converter("SklearnIncrementalPCA", convert_truncated_svd)
register_converter("SklearnPCA", convert_truncated_svd)
register_converter("SklearnTruncatedSVD", convert_truncated_svd)



# SPDX-License-Identifier: Apache-2.0


import numbers
from ..common._registration import register_shape_calculator
from ..common.data_types import FloatTensorType, Int64TensorType, DoubleTensorType
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_scaler_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C_1], ..., [N, C_n] ---> [N, C_1 + ... + C_n]

    Similar to imputer, this operator can take multiple input feature
    tensors and concatenate them along C-axis.
    """
    check_input_and_output_numbers(
        operator, input_count_range=[1, None], output_count_range=1
    )
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType],
    )
    # Inputs: multiple float- and integer-tensors
    # Output: one float tensor
    for variable in operator.inputs:
        if len(set(variable.get_first_dimension() for variable in operator.inputs)) > 1:
            raise RuntimeError("Batch size must be identical across inputs.")

    N = operator.inputs[0].get_first_dimension()
    C = 0
    for variable in operator.inputs:
        c = variable.get_second_dimension()
        if isinstance(c, numbers.Integral):
            C += c
        else:
            C = None
            break

    operator.outputs[0].type.shape = [N, C]


register_shape_calculator("SklearnRobustScaler", calculate_sklearn_scaler_output_shapes)
register_shape_calculator("SklearnScaler", calculate_sklearn_scaler_output_shapes)
register_shape_calculator("SklearnNormalizer", calculate_sklearn_scaler_output_shapes)
register_shape_calculator("SklearnMinMaxScaler", calculate_sklearn_scaler_output_shapes)
register_shape_calculator("SklearnMaxAbsScaler", calculate_sklearn_scaler_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import numpy as np

try:
    from onnx.helper import np_dtype_to_tensor_dtype
except ImportError:
    from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE

    def np_dtype_to_tensor_dtype(dtype):
        return NP_TYPE_TO_TENSOR_TYPE[dtype]


from onnxconverter_common.onnx_ops import *  # noqa
from ..proto import onnx_proto


def apply_normalizer(scope, inputs, outputs, container, norm, use_float):
    """
    Adds operator Normalizer if *use_float* is true,
    otherwise, uses *ReduceSum* + *Div*. *Normalizer*
    always produces float according to ONNX speciciations.
    """
    input = inputs[0] if isinstance(inputs, list) else inputs
    output = outputs[0] if isinstance(outputs, list) else outputs
    use_normalizer = container.is_allowed({"Normalizer"})

    if use_normalizer and use_float:
        container.add_node(
            "Normalizer",
            input,
            output,
            op_domain="ai.onnx.ml",
            norm=norm,
            name=scope.get_unique_operator_name("Normalizer"),
        )
    else:
        # Normalizer only produces floats.
        if norm == "L1":
            norm = scope.get_unique_variable_name("norm")
            norm_abs = scope.get_unique_variable_name("norm_abs")
            container.add_node(
                "Abs", input, norm_abs, name=scope.get_unique_operator_name("Abs")
            )

            if container.target_opset < 13:
                container.add_node(
                    "ReduceSum",
                    norm_abs,
                    norm,
                    axes=[1],
                    keepdims=1,
                    name=scope.get_unique_operator_name("ReduceSum"),
                )
            else:
                axis_name = scope.get_unique_variable_name("axis")
                container.add_initializer(
                    axis_name, onnx_proto.TensorProto.INT64, [1], [1]
                )
                container.add_node(
                    "ReduceSum",
                    [norm_abs, axis_name],
                    norm,
                    keepdims=1,
                    name=scope.get_unique_operator_name("ReduceSum"),
                )
            apply_div(  # noqa
                scope,
                [input, norm],
                output,
                container,
                operator_name=scope.get_unique_operator_name("NormalizerNorm"),
            )
        elif norm == "L2":
            norm = scope.get_unique_variable_name("norm")
            norm2 = scope.get_unique_variable_name("norm2")
            if container.target_opset < 18:
                container.add_node(
                    "ReduceSumSquare",
                    input,
                    norm,
                    axes=[1],
                    keepdims=1,
                    name=scope.get_unique_operator_name("ReduceSumSquare"),
                )
            else:
                axis_name = scope.get_unique_variable_name("axis")
                container.add_initializer(
                    axis_name, onnx_proto.TensorProto.INT64, [1], [1]
                )
                container.add_node(
                    "ReduceSumSquare",
                    [input, axis_name],
                    norm,
                    keepdims=1,
                    name=scope.get_unique_operator_name("ReduceSumSquare"),
                )
            container.add_node(
                "Sqrt", [norm], norm2, name=scope.get_unique_operator_name("Sqrt")
            )
            apply_div(  # noqa
                scope,
                [input, norm2],
                output,
                container,
                operator_name=scope.get_unique_operator_name("NormalizerNorm"),
            )
        else:
            raise NotImplementedError(
                "Normalization not implemented for norm %r." % norm
            )


def _create_name_or_use_existing_one(scope, op_type, name):
    if name is None:
        return scope.get_unique_operator_name(op_type)
    return name


def apply_clip(
    scope, input_name, output_name, container, operator_name=None, max=None, min=None
):
    name = _create_name_or_use_existing_one(scope, "Clip", operator_name)
    attrs = {"name": name}

    if container.target_opset < 11:
        if max is not None:
            attrs["max"] = float(max)
        if min is not None:
            attrs["min"] = float(min)

        if container.target_opset < 6:
            attrs["consumed_inputs"] = [0]
            op_version = 1
        else:
            op_version = 6

        container.add_node(
            "Clip", input_name, output_name, op_version=op_version, **attrs
        )
    else:
        if container.target_opset < 12:
            op_version = 11
        else:
            op_version = 12
        if min is None and max is not None:
            raise RuntimeError("Operator 'Clip': min must be specified if max is.")
        inputs = [input_name]

        if min is not None:
            if isinstance(
                min,
                (np.ndarray, float, int, np.float32, np.float64, np.int64, np.int32),
            ):
                # add initializer
                if isinstance(min, np.ndarray):
                    if len(min.shape) == 0:
                        min = [min]
                    elif min.shape == (1,):
                        min = list(min[0]) if hasattr(min[0], "__iter__") else list(min)
                    else:
                        raise RuntimeError("min must be an array of one element.")
                else:
                    min = [min]

                # container in sklearn-onnx stores the computation type in
                # container.dtype.
                min_name = scope.get_unique_variable_name("clip_min")
                if op_version < 12:
                    min = np.array(min, dtype=getattr(container, "dtype", np.float32))
                    container.add_initializer(
                        min_name,
                        getattr(container, "proto_dtype", onnx_proto.TensorProto.FLOAT),
                        [],
                        [min[0]],
                    )
                else:
                    min = np.array(min)
                    container.add_initializer(
                        min_name, np_dtype_to_tensor_dtype(min.dtype), [], [min[0]]
                    )
                min = min_name
            if isinstance(min, str):
                inputs.append(min)
            else:
                raise RuntimeError("Parameter 'min' must be a string or a float.")

        if max is not None:
            if min is None:
                raise RuntimeError("Parameter 'min' must be specified if 'max' is.")
            if isinstance(
                max,
                (np.ndarray, float, int, np.float32, np.float64, np.int64, np.int32),
            ):
                # add initializer
                if isinstance(max, np.ndarray):
                    if len(max.shape) == 0:
                        max = [max]
                    elif max.shape == (1,):
                        max = list(max[0]) if hasattr(max[0], "__iter__") else list(max)
                    else:
                        raise RuntimeError("max must be an array of one element.")
                else:
                    max = [max]

                max_name = scope.get_unique_variable_name("clip_max")
                if op_version < 12:
                    max = np.array(max, dtype=getattr(container, "dtype", np.float32))
                    container.add_initializer(
                        max_name,
                        getattr(container, "proto_dtype", onnx_proto.TensorProto.FLOAT),
                        [],
                        [max[0]],
                    )
                else:
                    max = np.array(max)
                    container.add_initializer(
                        max_name, np_dtype_to_tensor_dtype(max.dtype), [], [max[0]]
                    )
                max = max_name
            if isinstance(max, str):
                inputs.append(max)
            else:
                raise RuntimeError("Parameter 'max' must be a string or a float.")

        container.add_node("Clip", inputs, output_name, op_version=op_version, **attrs)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from onnxconverter_common.data_types import (  # noqa
    DataType,
    Int64Type,
    FloatType,  # noqa
    StringType,
    TensorType,  # noqa
    Int64TensorType,
    Int32TensorType,
    BooleanTensorType,  # noqa
    FloatTensorType,
    StringTensorType,
    DoubleTensorType,  # noqa
    DictionaryType,
    SequenceType,
)  # noqa

try:
    from onnxconverter_common.data_types import (  # noqa
        Complex64TensorType,
        Complex128TensorType,
    )
except ImportError:
    Complex64TensorType = None
    Complex128TensorType = None
from ..proto import TensorProto, onnx_proto


try:
    from onnxconverter_common.data_types import DoubleType
except ImportError:

    class DoubleType(DataType):
        def __init__(self, doc_string=""):
            super(DoubleType, self).__init__([1, 1], doc_string)

        def to_onnx_type(self):
            onnx_type = onnx_proto.TypeProto()
            onnx_type.tensor_type.elem_type = onnx_proto.TensorProto.DOUBLE
            s = onnx_type.tensor_type.shape.dim.add()
            s.dim_value = 1
            return onnx_type

        def __repr__(self):
            return "{}()".format(self.__class__.__name__)


try:
    from onnxconverter_common.data_types import Float16TensorType
except ImportError:

    class Float16TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(Float16TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.FLOAT16


try:
    from onnxconverter_common.data_types import Int8TensorType
except ImportError:

    class Int8TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(Int8TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.INT8


try:
    from onnxconverter_common.data_types import Int16TensorType
except ImportError:

    class Int16TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(Int16TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.INT16


try:
    from onnxconverter_common.data_types import UInt16TensorType
except ImportError:

    class UInt16TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(UInt16TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.UINT16


try:
    from onnxconverter_common.data_types import UInt32TensorType
except ImportError:

    class UInt32TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(UInt32TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.UINT32


try:
    from onnxconverter_common.data_types import UInt64TensorType
except ImportError:

    class UInt64TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(UInt64TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.UINT64


try:
    from onnxconverter_common.data_types import UInt8TensorType
except ImportError:

    class UInt8TensorType(TensorType):
        def __init__(self, shape=None, doc_string=""):
            super(UInt8TensorType, self).__init__(shape, doc_string)

        def _get_element_onnx_type(self):
            return onnx_proto.TensorProto.UINT8


try:
    from onnxconverter_common.data_types import UInt8Type
except ImportError:

    class UInt8Type(DataType):
        def __init__(self, doc_string=""):
            super(UInt8Type, self).__init__([1, 1], doc_string)

        def to_onnx_type(self):
            onnx_type = onnx_proto.TypeProto()
            onnx_type.tensor_type.elem_type = onnx_proto.TensorProto.UINT8
            s = onnx_type.tensor_type.shape.dim.add()
            s.dim_value = 1
            return onnx_type

        def __repr__(self):
            return "{}()".format(self.__class__.__name__)


try:
    from onnxconverter_common.data_types import Int8Type
except ImportError:

    class Int8Type(DataType):
        def __init__(self, doc_string=""):
            super(Int8Type, self).__init__([1, 1], doc_string)

        def to_onnx_type(self):
            onnx_type = onnx_proto.TypeProto()
            onnx_type.tensor_type.elem_type = onnx_proto.TensorProto.INT8
            s = onnx_type.tensor_type.shape.dim.add()
            s.dim_value = 1
            return onnx_type

        def __repr__(self):
            return "{}()".format(self.__class__.__name__)


def copy_type(vtype, empty=True):
    if isinstance(vtype, SequenceType):
        return vtype.__class__(copy_type(vtype.element_type))
    if isinstance(vtype, DictionaryType):
        return vtype.__class__(copy_type(vtype.key_type), copy_type(vtype.value_type))
    return vtype.__class__()


def _guess_type_proto(data_type, dims):
    # This could be moved to onnxconverter_common.
    for d in dims:
        if d == 0:
            raise RuntimeError("Dimension should not be null: {}.".format(list(dims)))
    if data_type == onnx_proto.TensorProto.FLOAT:
        return FloatTensorType(dims)
    if data_type == onnx_proto.TensorProto.DOUBLE:
        return DoubleTensorType(dims)
    if data_type == onnx_proto.TensorProto.STRING:
        return StringTensorType(dims)
    if data_type == onnx_proto.TensorProto.INT64:
        return Int64TensorType(dims)
    if data_type == onnx_proto.TensorProto.INT32:
        return Int32TensorType(dims)
    if data_type == onnx_proto.TensorProto.BOOL:
        return BooleanTensorType(dims)
    if data_type == onnx_proto.TensorProto.INT8:
        return Int8TensorType(dims)
    if data_type == onnx_proto.TensorProto.UINT8:
        return UInt8TensorType(dims)
    if Complex64TensorType is not None:
        if data_type == onnx_proto.TensorProto.COMPLEX64:
            return Complex64TensorType(dims)
        if data_type == onnx_proto.TensorProto.COMPLEX128:
            return Complex128TensorType(dims)
    raise NotImplementedError(
        "Unsupported data_type '{}'. You may raise an issue "
        "at https://github.com/onnx/sklearn-onnx/issues."
        "".format(data_type)
    )


def _guess_type_proto_str(data_type, dims):
    # This could be moved to onnxconverter_common.
    if data_type == "tensor(float)":
        return FloatTensorType(dims)
    if data_type == "tensor(double)":
        return DoubleTensorType(dims)
    if data_type == "tensor(string)":
        return StringTensorType(dims)
    if data_type == "tensor(int64)":
        return Int64TensorType(dims)
    if data_type == "tensor(int32)":
        return Int32TensorType(dims)
    if data_type == "tensor(bool)":
        return BooleanTensorType(dims)
    if data_type == "tensor(int8)":
        return Int8TensorType(dims)
    if data_type == "tensor(uint8)":
        return UInt8TensorType(dims)
    if Complex64TensorType is not None:
        if data_type == "tensor(complex64)":
            return Complex64TensorType(dims)
        if data_type == "tensor(complex128)":
            return Complex128TensorType(dims)
    raise NotImplementedError(
        "Unsupported data_type '{}'. You may raise an issue "
        "at https://github.com/onnx/sklearn-onnx/issues."
        "".format(data_type)
    )


def _guess_type_proto_str_inv(data_type):
    # This could be moved to onnxconverter_common.
    if isinstance(data_type, FloatTensorType):
        return "tensor(float)"
    if isinstance(data_type, DoubleTensorType):
        return "tensor(double)"
    if isinstance(data_type, StringTensorType):
        return "tensor(string)"
    if isinstance(data_type, Int64TensorType):
        return "tensor(int64)"
    if isinstance(data_type, Int32TensorType):
        return "tensor(int32)"
    if isinstance(data_type, BooleanTensorType):
        return "tensor(bool)"
    raise NotImplementedError(
        "Unsupported data_type '{}'. You may raise an issue "
        "at https://github.com/onnx/sklearn-onnx/issues."
        "".format(data_type)
    )


def _guess_numpy_type(data_type, dims):
    # This could be moved to onnxconverter_common.
    if data_type == np.float32:
        return FloatTensorType(dims)
    if data_type == np.float64:
        return DoubleTensorType(dims)
    if (
        data_type in (np.str_, str, object)
        or str(data_type) in ("<U1",)
        or (hasattr(data_type, "type") and data_type.type is np.str_)
    ):  # noqa
        return StringTensorType(dims)
    if data_type in (np.int64,) or str(data_type) == "<U6":
        return Int64TensorType(dims)
    if data_type in (np.int32,) or str(data_type) in ("<U4",):  # noqa
        return Int32TensorType(dims)
    if data_type == np.uint8:
        return UInt8TensorType(dims)
    if data_type in (np.bool_, bool):
        return BooleanTensorType(dims)
    if data_type in (np.str_, str):
        return StringTensorType(dims)
    if data_type == np.int8:
        return Int8TensorType(dims)
    if data_type == np.int16:
        return Int16TensorType(dims)
    if data_type == np.uint64:
        return UInt64TensorType(dims)
    if data_type == np.uint32:
        return UInt32TensorType(dims)
    if data_type == np.uint16:
        return UInt16TensorType(dims)
    if data_type == np.float16:
        return Float16TensorType(dims)
    if Complex64TensorType is not None:
        if data_type == np.complex64:
            return Complex64TensorType(dims)
        if data_type == np.complex128:
            return Complex128TensorType(dims)
    raise NotImplementedError(
        "Unsupported data_type %r (type=%r). You may raise an issue "
        "at https://github.com/onnx/sklearn-onnx/issues."
        "" % (data_type, type(data_type))
    )


def guess_data_type(type_, shape=None):
    """
    Guess the datatype given the type type_
    """
    if isinstance(type_, TensorProto):
        return _guess_type_proto(type, shape)
    if isinstance(type_, str):
        return _guess_type_proto_str(type_, shape)
    if hasattr(type_, "columns") and hasattr(type_, "dtypes"):
        # DataFrame
        return [
            (name, _guess_numpy_type(dt, [None, 1]))
            for name, dt in zip(type_.columns, type_.dtypes)
        ]
    if hasattr(type_, "name") and hasattr(type_, "dtype"):
        # Series
        return [(type_.name, _guess_numpy_type(type_.dtype, [None, 1]))]
    if hasattr(type_, "shape") and hasattr(type_, "dtype"):
        # array
        return [("input", _guess_numpy_type(type_.dtype, type_.shape))]
    raise TypeError(
        "Type {} cannot be converted into a "
        "DataType. You may raise an issue at "
        "https://github.com/onnx/sklearn-onnx/issues."
        "".format(type(type_))
    )


def guess_numpy_type(data_type):
    """
    Guess the corresponding numpy type based on data_type.
    """
    if data_type in (
        np.float64,
        np.float32,
        np.int8,
        np.uint8,
        np.str_,
        np.bool_,
        np.int32,
        np.int64,
    ):
        return data_type
    if data_type == str:
        return np.str_
    if data_type == bool:
        return np.bool_
    if isinstance(data_type, FloatTensorType):
        return np.float32
    if isinstance(data_type, DoubleTensorType):
        return np.float64
    if isinstance(data_type, Int32TensorType):
        return np.int32
    if isinstance(data_type, Int64TensorType):
        return np.int64
    if isinstance(data_type, StringTensorType):
        return np.str_
    if isinstance(data_type, BooleanTensorType):
        return np.bool_
    if Complex64TensorType is not None:
        if data_type in (np.complex64, np.complex128):
            return data_type
        if isinstance(data_type, Complex64TensorType):
            return np.complex64
        if isinstance(data_type, Complex128TensorType):
            return np.complex128
    raise NotImplementedError("Unsupported data_type '{}'.".format(data_type))


def guess_proto_type(data_type):
    """
    Guess the corresponding proto type based on data_type.
    """
    if isinstance(data_type, FloatTensorType):
        return onnx_proto.TensorProto.FLOAT
    if isinstance(data_type, DoubleTensorType):
        return onnx_proto.TensorProto.DOUBLE
    if isinstance(data_type, Int32TensorType):
        return onnx_proto.TensorProto.INT32
    if isinstance(data_type, Int64TensorType):
        return onnx_proto.TensorProto.INT64
    if isinstance(data_type, StringTensorType):
        return onnx_proto.TensorProto.STRING
    if isinstance(data_type, BooleanTensorType):
        return onnx_proto.TensorProto.BOOL
    if isinstance(data_type, Int8TensorType):
        return onnx_proto.TensorProto.INT8
    if isinstance(data_type, UInt8TensorType):
        return onnx_proto.TensorProto.UINT8
    if Complex64TensorType is not None:
        if isinstance(data_type, Complex64TensorType):
            return onnx_proto.TensorProto.COMPLEX64
        if isinstance(data_type, Complex128TensorType):
            return onnx_proto.TensorProto.COMPLEX128
    raise NotImplementedError("Unsupported data_type '{}'.".format(data_type))


def guess_tensor_type(data_type):
    """
    Guess the corresponding variable output type based
    on input type. It returns type if *data_type* is a real.
    It returns *FloatTensorType* if *data_type* is an integer.
    """
    if isinstance(data_type, DoubleTensorType):
        return DoubleTensorType()
    if isinstance(data_type, DictionaryType):
        return guess_tensor_type(data_type.value_type)
    if Complex64TensorType is not None:
        if isinstance(data_type, (Complex64TensorType, Complex128TensorType)):
            return data_type.__class__()
    if not isinstance(
        data_type,
        (
            Int64TensorType,
            Int32TensorType,
            BooleanTensorType,
            FloatTensorType,
            StringTensorType,
            DoubleTensorType,
            Int8TensorType,
            UInt8TensorType,
        ),
    ):
        raise TypeError(
            "data_type is not a tensor type but '{}'.".format(type(data_type))
        )
    return FloatTensorType()



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_label_encoder(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    op_type = "LabelEncoder"
    attrs = {"name": scope.get_unique_operator_name(op_type)}
    classes = op.classes_
    if np.issubdtype(classes.dtype, np.floating):
        attrs["keys_floats"] = classes
    elif np.issubdtype(classes.dtype, np.signedinteger) or classes.dtype == np.bool_:
        attrs["keys_int64s"] = [int(i) for i in classes]
    else:
        attrs["keys_strings"] = np.array([s.encode("utf-8") for s in classes])
    attrs["values_int64s"] = np.arange(len(classes))

    cop = container.target_opset_any_domain("ai.onnx.ml")
    if cop is not None and cop < 2:
        raise RuntimeError(
            "LabelEncoder requires at least opset 2 for domain 'ai.onnx.ml' "
            "not {}".format(cop)
        )

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        op_version=2,
        **attrs
    )


register_converter("SklearnLabelEncoder", convert_sklearn_label_encoder)



# SPDX-License-Identifier: Apache-2.0
import sys
from textwrap import dedent


def _help():
    print(
        dedent(
            """
        python -m skl2onnx [command]

    command is:

    setup   generate rst documentation for every ONNX operator
            before building the package"""
        )
    )


def _setup():
    from skl2onnx.algebra.onnx_ops import dynamic_class_creation

    dynamic_class_creation(True)


def main(argv):
    if len(argv) <= 1 or "--help" in argv:
        _help()
        return

    if "setup" in argv:
        print("generate rst documentation for every ONNX operator")
        _setup()
        return

    _help()


if __name__ == "__main__":
    main(sys.argv)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from onnx.numpy_helper import from_array
from ..common._apply_operation import (
    apply_cast,
    apply_concat,
    apply_div,
    apply_mul,
    apply_reshape,
    apply_reducesum,
    apply_transpose,
)
from ..common._registration import register_converter
from ..common.data_types import (
    BooleanTensorType,
    Int64TensorType,
    guess_numpy_type,
    guess_proto_type,
)
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    get_default_tree_classifier_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto


def populate_tree_attributes(model, name, dtype):
    """Construct attrs dictionary to be used in predict()
    while adding a node with TreeEnsembleClassifier ONNX op.
    """
    attrs = {}
    attrs["name"] = name
    attrs["post_transform"] = "NONE"
    attrs["nodes_treeids"] = []
    attrs["nodes_nodeids"] = []
    attrs["nodes_featureids"] = []
    attrs["nodes_modes"] = []
    attrs["nodes_values"] = []
    attrs["nodes_truenodeids"] = []
    attrs["nodes_falsenodeids"] = []
    attrs["nodes_missing_value_tracks_true"] = []
    attrs["nodes_hitrates"] = []
    attrs["class_treeids"] = []
    attrs["class_nodeids"] = []
    attrs["class_ids"] = []
    attrs["class_weights"] = []
    attrs["classlabels_int64s"] = list(range(model.tree_.node_count))

    for i in range(model.tree_.node_count):
        node_id = i
        if model.tree_.children_left[i] > i and model.tree_.children_right[i] > i:
            feat = model.tree_.feature[i]
            thresh = model.tree_.threshold[i]
            left = model.tree_.children_left[i]
            right = model.tree_.children_right[i]
            mode = "BRANCH_LEQ"
        else:
            feat, thresh, left, right = 0, 0.0, 0, 0
            mode = "LEAF"
        attrs["nodes_nodeids"].append(node_id)
        attrs["nodes_treeids"].append(0)
        attrs["nodes_featureids"].append(feat)
        attrs["nodes_modes"].append(mode)
        attrs["nodes_truenodeids"].append(left)
        attrs["nodes_falsenodeids"].append(right)
        attrs["nodes_missing_value_tracks_true"].append(False)
        attrs["nodes_hitrates"].append(1.0)
        attrs["nodes_values"].append(thresh)
        if mode == "LEAF":
            attrs["class_ids"].append(node_id)
            attrs["class_weights"].append(1.0)
            attrs["class_treeids"].append(0)
            attrs["class_nodeids"].append(node_id)
    if dtype is not None:
        for k in attrs:
            if k in ("node_values", "class_weights", "target_weights"):
                attrs[k] = np.array(attrs[k], dtype=dtype)
    return attrs


def predict(
    model, scope, operator, container, op_type, op_domain, op_version, is_ensemble=False
):
    """Predict target and calculate probability scores."""
    indices_name = scope.get_unique_variable_name("indices")
    dummy_proba_name = scope.get_unique_variable_name("dummy_proba")
    values_name = scope.get_unique_variable_name("values")
    out_values_name = scope.get_unique_variable_name("out_indices")
    transposed_result_name = scope.get_unique_variable_name("transposed_result")
    proba_output_name = scope.get_unique_variable_name("proba_output")
    cast_result_name = scope.get_unique_variable_name("cast_result")
    reshaped_indices_name = scope.get_unique_variable_name("reshaped_indices")
    sum_output_name = scope.get_unique_variable_name("sum_proba")
    value = model.tree_.value.transpose(1, 2, 0)

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    container.add_initializer(values_name, proto_dtype, value.shape, value.ravel())

    input_name = operator.input_full_names
    if isinstance(operator.inputs[0].type, BooleanTensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(scope, input_name, cast_input_name, container, to=proto_dtype)
        input_name = cast_input_name

    if model.tree_.node_count > 1:
        attrs = populate_tree_attributes(
            model, scope.get_unique_operator_name(op_type), dtype
        )
        container.add_node(
            op_type,
            input_name,
            [indices_name, dummy_proba_name],
            op_domain=op_domain,
            op_version=op_version,
            **attrs,
        )
    else:
        zero_name = scope.get_unique_variable_name("zero")
        zero_matrix_name = scope.get_unique_variable_name("zero_matrix")
        reduced_zero_matrix_name = scope.get_unique_variable_name("reduced_zero_matrix")

        container.add_initializer(zero_name, proto_dtype, [], [0])
        apply_mul(
            scope, [input_name[0], zero_name], zero_matrix_name, container, broadcast=1
        )
        if container.target_opset < 13:
            container.add_node(
                "ReduceSum",
                zero_matrix_name,
                reduced_zero_matrix_name,
                axes=[1],
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        else:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                "ReduceSum",
                [zero_matrix_name, axis_name],
                reduced_zero_matrix_name,
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        apply_cast(
            scope,
            reduced_zero_matrix_name,
            indices_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    apply_reshape(
        scope, indices_name, reshaped_indices_name, container, desired_shape=[1, -1]
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [values_name, reshaped_indices_name],
        out_values_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )
    apply_transpose(
        scope, out_values_name, proba_output_name, container, perm=(0, 2, 1)
    )

    if is_ensemble:
        proba_result_name = scope.get_unique_variable_name("proba_result")
        apply_reducesum(
            scope, proba_output_name, sum_output_name, container, keepdims=1, axes=[2]
        )
        apply_div(
            scope, [proba_output_name, sum_output_name], proba_result_name, container
        )
        return proba_result_name
    else:
        apply_cast(
            scope,
            proba_output_name,
            cast_result_name,
            container,
            to=onnx_proto.TensorProto.BOOL,
        )
        apply_cast(
            scope,
            cast_result_name,
            operator.outputs[1].full_name,
            container,
            to=proto_dtype,
        )
        apply_transpose(
            scope, out_values_name, transposed_result_name, container, perm=(2, 1, 0)
        )
        return transposed_result_name


def _append_decision_output(
    input_name,
    attrs,
    fct_label,
    n_out,
    scope,
    operator,
    container,
    op_type="TreeEnsembleClassifier",
    op_domain="ai.onnx.ml",
    op_version=1,
    cast_encode=False,
    regression=False,
    dtype=np.float32,
    overwrite_tree=None,
):
    attrs = attrs.copy()
    attrs["name"] = scope.get_unique_operator_name(op_type)
    attrs["n_targets"] = 1
    attrs["post_transform"] = "NONE"
    if regression:
        attrs["target_weights"] = np.array(
            [float(_) for _ in attrs["target_nodeids"]], dtype=dtype
        )
    else:
        attrs["target_ids"] = [0 for _ in attrs["class_ids"]]
        attrs["target_weights"] = [float(_) for _ in attrs["class_nodeids"]]
        attrs["target_nodeids"] = attrs["class_nodeids"]
        attrs["target_treeids"] = attrs["class_treeids"]

    rem = [k for k in attrs if k.startswith("class")]
    for k in rem:
        del attrs[k]
    dpath = scope.get_unique_variable_name("dpath")
    container.add_node(
        op_type.replace("Classifier", "Regressor"),
        input_name,
        dpath,
        op_domain=op_domain,
        op_version=op_version,
        **attrs,
    )

    if n_out is None:
        final_name = scope.get_unique_variable_name("dpatho")
    else:
        final_name = operator.outputs[n_out].full_name

    if cast_encode:
        apply_cast(
            scope,
            dpath,
            final_name,
            container,
            to=onnx_proto.TensorProto.INT64,
            operator_name=scope.get_unique_operator_name("TreePathType"),
        )
    else:
        op = operator.raw_operator
        labels = fct_label(overwrite_tree if overwrite_tree is not None else op.tree_)
        ordered = list(sorted(labels.items()))
        keys = [float(_[0]) for _ in ordered]
        values = [_[1] for _ in ordered]
        name = scope.get_unique_variable_name("spath")
        container.add_node(
            "LabelEncoder",
            dpath,
            name,
            op_domain=op_domain,
            op_version=2,
            default_string="0",
            keys_floats=keys,
            values_strings=values,
            name=scope.get_unique_operator_name("TreePath"),
        )
        apply_reshape(
            scope,
            name,
            final_name,
            container,
            desired_shape=(-1, 1),
            operator_name=scope.get_unique_operator_name("TreePathShape"),
        )

    return final_name


def convert_sklearn_decision_tree_classifier(
    scope,
    operator,
    container,
    op_type="TreeEnsembleClassifier",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    try:
        dtype = guess_numpy_type(operator.inputs[0].type)
    except NotImplementedError as e:
        raise RuntimeError("Unknown variable {}.".format(operator.inputs[0])) from e
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator
    options = scope.get_options(op, dict(decision_path=False, decision_leaf=False))
    if np.asarray(op.classes_).size == 1:
        # The model was trained with one label.
        # There is no need to build a tree.
        if op.n_outputs_ != 1:
            raise RuntimeError(
                f"One training class and multiple outputs is not "
                f"supported yet for class {op.__class__.__name__!r}."
            )
        if options["decision_path"] or options["decision_leaf"]:
            raise RuntimeError(
                f"One training class, option 'decision_path' "
                f"or 'decision_leaf' are not supported for "
                f"class {op.__class__.__name__!r}."
            )

        zero = scope.get_unique_variable_name("zero")
        one = scope.get_unique_variable_name("one")
        new_shape = scope.get_unique_variable_name("new_shape")
        container.add_initializer(zero, onnx_proto.TensorProto.INT64, [1], [0])
        container.add_initializer(one, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_initializer(new_shape, onnx_proto.TensorProto.INT64, [2], [-1, 1])
        shape = scope.get_unique_variable_name("shape")
        container.add_node("Shape", [operator.inputs[0].full_name], [shape])
        shape_sliced = scope.get_unique_variable_name("shape_sliced")
        container.add_node("Slice", [shape, zero, one, zero], [shape_sliced])

        # labels
        container.add_node(
            "ConstantOfShape",
            [shape_sliced],
            [operator.outputs[0].full_name],
            value=from_array(np.array([op.classes_[0]], dtype=np.int64)),
        )

        # probabilities
        probas = scope.get_unique_variable_name("probas")
        container.add_node(
            "ConstantOfShape",
            [shape_sliced],
            [probas],
            value=from_array(np.array([1], dtype=dtype)),
        )
        container.add_node(
            "Reshape", [probas, new_shape], [operator.outputs[1].full_name]
        )
        return

    if op.n_outputs_ == 1:
        attrs = get_default_tree_classifier_attribute_pairs()
        attrs["name"] = scope.get_unique_operator_name(op_type)
        classes = get_label_classes(scope, op)

        if all(isinstance(i, np.ndarray) for i in classes):
            classes = np.concatenate(classes)
        if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
            class_labels = [int(i) for i in classes]
            attrs["classlabels_int64s"] = class_labels
        elif all(isinstance(i, str) for i in classes):
            class_labels = [str(i) for i in classes]
            attrs["classlabels_strings"] = class_labels
        else:
            raise ValueError("Labels must be all integers or all strings.")

        add_tree_to_attribute_pairs(
            attrs, True, op.tree_, 0, 1.0, 0, True, True, dtype=dtype
        )
        input_name = operator.input_full_names
        if isinstance(operator.inputs[0].type, BooleanTensorType):
            cast_input_name = scope.get_unique_variable_name("cast_input")

            apply_cast(
                scope,
                input_name,
                cast_input_name,
                container,
                to=onnx_proto.TensorProto.FLOAT,
            )
            input_name = cast_input_name

        if dtype is not None:
            for k in attrs:
                if k in (
                    "nodes_values",
                    "class_weights",
                    "target_weights",
                    "nodes_hitrates",
                    "base_values",
                ):
                    attrs[k] = np.array(attrs[k], dtype=dtype)

        container.add_node(
            op_type,
            input_name,
            [operator.outputs[0].full_name, operator.outputs[1].full_name],
            op_domain=op_domain,
            op_version=op_version,
            **attrs,
        )

        n_out = 2
        if options["decision_path"]:
            # decision_path
            _append_decision_output(
                input_name,
                attrs,
                _build_labels_path,
                n_out,
                scope,
                operator,
                container,
                op_type=op_type,
                op_domain=op_domain,
                op_version=op_version,
                dtype=dtype,
            )
            n_out += 1
        if options["decision_leaf"]:
            # decision_path
            _append_decision_output(
                input_name,
                attrs,
                _build_labels_leaf,
                n_out,
                scope,
                operator,
                container,
                op_type=op_type,
                op_domain=op_domain,
                op_version=op_version,
                cast_encode=True,
                dtype=dtype,
            )
            n_out += 1
    else:
        transposed_result_name = predict(
            op, scope, operator, container, op_type, op_domain, op_version
        )
        predictions = []
        for k in range(op.n_outputs_):
            preds_name = scope.get_unique_variable_name("preds")
            reshaped_preds_name = scope.get_unique_variable_name("reshaped_preds")
            k_name = scope.get_unique_variable_name("k_column")
            out_k_name = scope.get_unique_variable_name("out_k_column")
            argmax_output_name = scope.get_unique_variable_name("argmax_output")
            classes_name = scope.get_unique_variable_name("classes")
            reshaped_result_name = scope.get_unique_variable_name("reshaped_result")

            container.add_initializer(k_name, onnx_proto.TensorProto.INT64, [], [k])
            container.add_initializer(
                classes_name,
                onnx_proto.TensorProto.INT64,
                op.classes_[k].shape,
                [int(i) for i in op.classes_[k]],
            )

            container.add_node(
                "ArrayFeatureExtractor",
                [transposed_result_name, k_name],
                out_k_name,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
            )
            container.add_node(
                "ArgMax",
                out_k_name,
                argmax_output_name,
                name=scope.get_unique_operator_name("ArgMax"),
                axis=1,
            )
            apply_reshape(
                scope,
                argmax_output_name,
                reshaped_result_name,
                container,
                desired_shape=(1, -1),
            )
            container.add_node(
                "ArrayFeatureExtractor",
                [classes_name, reshaped_result_name],
                preds_name,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
            )
            apply_reshape(
                scope, preds_name, reshaped_preds_name, container, desired_shape=(-1, 1)
            )
            predictions.append(reshaped_preds_name)
        apply_concat(
            scope, predictions, operator.outputs[0].full_name, container, axis=1
        )

        if options["decision_path"]:
            raise RuntimeError(
                "Option decision_path for multi-outputs " "is not implemented yet."
            )
        if options["decision_leaf"]:
            raise RuntimeError(
                "Option decision_leaf for multi-outputs " "is not implemented yet."
            )


def convert_sklearn_decision_tree_regressor(
    scope,
    operator,
    container,
    op_type="TreeEnsembleRegressor",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator

    attrs = get_default_tree_regressor_attribute_pairs()
    attrs["name"] = scope.get_unique_operator_name(op_type)
    attrs["n_targets"] = int(op.n_outputs_)
    add_tree_to_attribute_pairs(
        attrs, False, op.tree_, 0, 1.0, 0, False, True, dtype=dtype
    )

    if dtype is not None:
        for k in attrs:
            if k in (
                "nodes_values",
                "class_weights",
                "target_weights",
                "nodes_hitrates",
                "base_values",
            ):
                attrs[k] = np.array(attrs[k], dtype=dtype)

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope,
            operator.input_full_names,
            cast_input_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        input_name = [cast_input_name]

    container.add_node(
        op_type,
        input_name,
        operator.outputs[0].full_name,
        op_domain=op_domain,
        op_version=op_version,
        **attrs,
    )

    options = scope.get_options(op, dict(decision_path=False, decision_leaf=False))

    # decision_path
    n_out = 1
    if options["decision_path"]:
        # decision_path
        _append_decision_output(
            input_name,
            attrs,
            _build_labels_path,
            n_out,
            scope,
            operator,
            container,
            op_type=op_type,
            op_domain=op_domain,
            op_version=op_version,
            regression=True,
        )
        n_out += 1
    if options["decision_leaf"]:
        # decision_path
        _append_decision_output(
            input_name,
            attrs,
            _build_labels_leaf,
            n_out,
            scope,
            operator,
            container,
            op_type=op_type,
            op_domain=op_domain,
            op_version=op_version,
            regression=True,
            cast_encode=True,
        )
        n_out += 1


def _recursive_build_labels(tree, index, current):
    current[index] = True
    if tree.children_left[index] == -1:
        yield (index, current.copy())
    else:
        for it in _recursive_build_labels(tree, tree.children_left[index], current):
            yield it
        for it in _recursive_build_labels(tree, tree.children_right[index], current):
            yield it
    current[index] = False


def _build_labels_path(tree):
    paths = {}
    current = {}

    for leave_index, path in _recursive_build_labels(tree, 0, current):
        spath = ["0" for _ in range(tree.node_count)]
        for nodeid, b in path.items():
            if b:
                spath[nodeid] = "1"
        paths[leave_index] = "".join(spath)
    return paths


def _build_labels_leaf(tree):
    paths = {}
    current = {}

    for leave_index, path in _recursive_build_labels(tree, 0, current):
        paths[leave_index] = leave_index
    return paths


register_converter(
    "SklearnDecisionTreeClassifier",
    convert_sklearn_decision_tree_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "decision_path": [True, False],
        "decision_leaf": [True, False],
    },
)
register_converter(
    "SklearnDecisionTreeRegressor",
    convert_sklearn_decision_tree_regressor,
    options={"decision_path": [True, False], "decision_leaf": [True, False]},
)
register_converter(
    "SklearnExtraTreeClassifier",
    convert_sklearn_decision_tree_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "decision_path": [True, False],
        "decision_leaf": [True, False],
    },
)
register_converter(
    "SklearnExtraTreeRegressor",
    convert_sklearn_decision_tree_regressor,
    options={"decision_path": [True, False], "decision_leaf": [True, False]},
)



# SPDX-License-Identifier: Apache-2.0

from sklearn.base import is_classifier
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_cast
from ..common.data_types import guess_proto_type
from .._parse import _parse_sklearn


def convert_pipeline(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    model = operator.raw_operator
    inputs = operator.inputs
    for step in model.steps:
        step_model = step[1]
        if is_classifier(step_model):
            scope.add_options(id(step_model), options={"zipmap": False})
            container.add_options(id(step_model), options={"zipmap": False})
        outputs = _parse_sklearn(scope, step_model, inputs, custom_parsers=None)
        inputs = outputs
    if len(outputs) != len(operator.outputs):
        raise RuntimeError(
            "Mismatch between pipeline output %d and "
            "last step outputs %d." % (len(outputs), len(operator.outputs))
        )
    for fr, to in zip(outputs, operator.outputs):
        if isinstance(to.type, type(fr.type)):
            container.add_node(
                "Identity",
                fr.full_name,
                to.full_name,
                name=scope.get_unique_operator_name("Id" + operator.onnx_name),
            )
        else:
            # If Pipeline output types are different with last stage output type
            apply_cast(
                scope,
                fr.full_name,
                to.full_name,
                container,
                operator_name=scope.get_unique_operator_name(
                    "Cast" + operator.onnx_name
                ),
                to=guess_proto_type(to.type),
            )


def convert_feature_union(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    raise NotImplementedError(
        "This converter not needed so far. It is usually handled " "during parsing."
    )


def convert_column_transformer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    raise NotImplementedError(
        "This converter not needed so far. It is usually handled " "during parsing."
    )


register_converter("SklearnPipeline", convert_pipeline)
register_converter("SklearnFeatureUnion", convert_feature_union)
register_converter("SklearnColumnTransformer", convert_column_transformer)
